"""This module contains utilities for DWI handling."""


def merge_volumes_tdim(in_file1, in_file2):
    """Merge 'in_file1' and 'in_file2' in the t dimension.

    Args:
        in_file1 (str): First set of volumes.
        in_file2 (str): Second set of volumes.

    Returns:
        out_file (str): The two sets of volumes merged.
    """
    import os

    out_file = os.path.abspath("merged_files.nii.gz")
    cmd = f"fslmerge -t {out_file} {in_file1} {in_file2}"
    os.system(cmd)
    return out_file


def count_b0s(in_bval, low_bval=5.0):
    """Count the number of volumes where b<=low_bval.

    Args:
        in_bval (str): bval file.
        low_bval (int, optional): Define the b0 volumes as all volume bval <= lowbval. Defaults to 5.0.

    Returns:
        num_b0s: Number of b0s.
    """
    import numpy as np

    bvals = np.loadtxt(in_bval)
    num_b0s = len(np.where(bvals <= low_bval)[0])

    return num_b0s


def b0_average(in_file, out_file=None):
    """
    Average the b0 volumes.

    Args:
        in_file (str): The b0 volumes already registered.
        out_file (str, optional): Name of the file. Defaults to None.

    Returns:
        The mean of the b0 volumes.

    Warnings:
        The b0 volumes must be registered.
    """
    import os

    import nibabel as nb
    import numpy as np

    if not out_file:
        fname, ext = os.path.splitext(os.path.basename(in_file))
        if ext == ".gz":
            fname, ext2 = os.path.splitext(fname)
            ext = ext2 + ext
        out_file = os.path.abspath(f"{fname}_avg_b0{ext}")

    imgs = np.array(nb.four_to_three(nb.load(in_file)))
    b0s = [im.get_fdata(dtype="float32").astype(np.float32) for im in imgs]
    b0 = np.average(np.array(b0s), axis=0)

    hdr = imgs[0].get_header().copy()
    hdr.set_data_shape(b0.shape)
    hdr.set_xyzt_units("mm")
    hdr.set_data_dtype(np.float32)
    nb.Nifti1Image(b0, imgs[0].get_affine(), hdr).to_filename(out_file)

    return out_file


def b0_dwi_split(in_dwi: str, in_bval: str, in_bvec: str, low_bval: float = 5.0):
    """Split DWI dataset.

    Split the DWI volumes into two datasets :
     - the first dataset contains the set of b<=low_bval volumes.
     - the second dataset contains the set of DWI volumes.

    Args:
        in_dwi (str): DWI dataset.
        in_bval (str): File describing the b-values of the DWI dataset.
        in_bvec (str): File describing the directions of the DWI dataset.
        low_bval (float, optional): Define the b0 volumes as all volume bval <= lowbval. Defaults to 5.0.

    Returns:
        out_b0 (str): The set of b<=low_bval volumes.
        out_dwi (str): Output. The set of b>low_bval volumes.
        out_bvals (str): The b-values corresponding to the out_dwi.
        out_bvecs (str): The b-vecs corresponding to the out_dwi.
    """
    import os
    import warnings

    import nibabel as nib
    import numpy as np

    assert os.path.isfile(in_dwi)
    assert os.path.isfile(in_bval)
    assert os.path.isfile(in_bvec)
    assert low_bval >= 0

    im = nib.load(in_dwi)
    data = im.get_fdata(dtype="float32")
    hdr = im.get_header().copy()
    bvals = np.loadtxt(in_bval)
    bvecs = np.loadtxt(in_bvec)

    if bvals.shape[0] == bvecs.shape[0]:
        warnings.warn(
            "Warning: The b-vectors file should be column-wise. The b-vectors will be transposed",
            UserWarning,
        )
        bvecs = bvecs.T

    lowbs = np.where(bvals <= low_bval)[0]

    fname_b0, ext_b0 = os.path.splitext(os.path.basename(in_dwi))
    if ext_b0 == ".gz":
        fname_b0, ext2 = os.path.splitext(fname_b0)
        ext_b0 = ext2 + ext_b0
    out_b0 = os.path.abspath(f"{fname_b0}_b0{ext_b0}")
    # out_b0 = op.abspath('b0.nii.gz')
    b0data = data[..., lowbs]
    hdr.set_data_shape(b0data.shape)
    nib.Nifti1Image(b0data, im.get_affine(), hdr).to_filename(out_b0)

    dwi_bvals = np.where(bvals > low_bval)[0]
    out_dwi = os.path.abspath("dwi.nii.gz")
    dwi_data = data[..., dwi_bvals]
    hdr.set_data_shape(dwi_data.shape)
    nib.Nifti1Image(dwi_data, im.get_affine(), hdr).to_filename(out_dwi)

    bvals_dwi = bvals[dwi_bvals]
    out_bvals = os.path.abspath("bvals")
    np.savetxt(out_bvals, bvals_dwi, fmt="%d", delimiter=" ")

    bvecs_dwi = np.array(
        [
            bvecs[0][dwi_bvals].tolist(),
            bvecs[1][dwi_bvals].tolist(),
            bvecs[2][dwi_bvals].tolist(),
        ]
    )
    out_bvecs = os.path.abspath("bvecs")
    np.savetxt(out_bvecs, bvecs_dwi, fmt="%10.5f", delimiter=" ")

    return out_b0, out_dwi, out_bvals, out_bvecs


def compute_average_b0(in_dwi: str, in_bval: str, low_bval: float = 5.0):
    """Compute average b0 volume from DWI dataset."""
    import os

    import nibabel
    import numpy as np

    assert os.path.isfile(in_dwi)
    assert os.path.isfile(in_bval)
    assert low_bval >= 0

    imgs = np.array(nibabel.four_to_three(nibabel.load(in_dwi)))
    bvals = np.loadtxt(in_bval)
    low_bvals = np.where(bvals <= low_bval)[0]

    fname_b0, ext_b0 = os.path.splitext(os.path.basename(in_dwi))
    if ext_b0 == ".gz":
        fname_b0, ext2 = os.path.splitext(fname_b0)
        ext_b0 = ext2 + ext_b0
    out_b0_average = os.path.abspath(f"{fname_b0}_avg_b0{ext_b0}")

    b0s_data = [imgs[i].get_data() for i in low_bvals]
    avg_b0_data = np.average(np.array(b0s_data), axis=0)

    hdr = imgs[0].get_header().copy()
    hdr.set_data_shape(avg_b0_data.shape)
    nibabel.Nifti1Image(avg_b0_data, imgs[0].get_affine(), hdr).to_filename(
        out_b0_average
    )

    return out_b0_average


def insert_b0_into_dwi(in_b0, in_dwi, in_bval, in_bvec):
    """Insert a b0 volume into the dwi dataset as the first volume and update the bvals and bvecs files.

    Args:
        in_b0 (str): One b=0 volume (could be the average of a b0 dataset).
        in_dwi (str): The set of DWI volumes.
        in_bval (str): File describing the b-values of the DWI dataset.
        in_bvec (str): File describing the directions of the DWI dataset.

    Returns:
        out_dwi (str): Diffusion dataset : b0 volume + dwi volumes.
        out_bval (str): B-values update.
        out_bvec (str): Directions of diffusion update.
    """
    import os

    import numpy as np

    assert os.path.isfile(in_b0)
    assert os.path.isfile(in_dwi)
    assert os.path.isfile(in_bval)
    assert os.path.isfile(in_bvec)

    out_dwi = merge_volumes_tdim(in_b0, in_dwi)

    lst = np.loadtxt(in_bval).tolist()
    lst.insert(0, 0)
    out_bvals = os.path.abspath("bvals")
    np.savetxt(out_bvals, np.matrix(lst), fmt="%d", delimiter=" ")

    bvecs = np.loadtxt(in_bvec)
    bvecs_0 = bvecs[0].tolist()
    bvecs_0.insert(0, 0.0)
    bvecs_1 = bvecs[1].tolist()
    bvecs_1.insert(0, 0.0)
    bvecs_2 = bvecs[2].tolist()
    bvecs_2.insert(0, 0.0)
    bvecs_dwi = np.array([bvecs_0, bvecs_1, bvecs_2])
    out_bvecs = os.path.abspath("bvecs")
    np.savetxt(out_bvecs, bvecs_dwi, fmt="%10.5f", delimiter=" ")

    return out_dwi, out_bvals, out_bvecs


def check_dwi_volume(in_dwi, in_bvec, in_bval):
    """Check that # DWI = # B-val = # B-vec.

    Raises
        ValueError if # DWI, # B-val and # B-vec mismatch.
    """
    import nibabel as nib
    import numpy as np

    bvals = np.loadtxt(in_bval)
    num_b_vals = len(bvals)

    bvecs = np.loadtxt(in_bvec)
    _, num_b_vecs = bvecs.shape

    img = nib.load(in_dwi)
    _, _, _, num_dwis = img.shape

    if not (num_b_vals == num_b_vecs == num_dwis):
        raise IOError(
            f"Number of DWIs, b-vals and b-vecs mismatch "
            f"(# DWI = {num_dwis}, # B-vec = {num_b_vecs}, #B-val = {num_b_vals}) "
        )


def generate_index_file(in_bval, low_bval=5.0, image_id=None):
    """Generate [`image_id`]_index.txt file for FSL eddy command.

    Args:
        in_bval (str): Bval file.
        low_bval (float): Define the b0 volumes as all volume bval <= low_bval. Default to 5.0.
        image_id (str, optional): Optional prefix. Defaults to None.

    Returns:
        out_index: [`image_id`]_index.txt or index.txt file.
    """
    import os

    import numpy as np

    assert os.path.isfile(in_bval)
    bvals = np.loadtxt(in_bval)
    idx_low_bvals = np.where(bvals <= low_bval)
    b0_index = idx_low_bvals[0].tolist()

    if not b0_index:
        raise ValueError(
            f"Could not find b-value <= {low_bval} in bval file ({in_bval}). Found values: {bvals}"
        )

    if image_id:
        out_index = os.path.abspath(f"{image_id}_index.txt")
    else:
        out_index = os.path.abspath("index.txt")

    vols = len(bvals)
    index_list = []
    for i in range(0, len(b0_index)):
        if i == (len(b0_index) - 1):
            index_list.extend([i + 1] * (vols - b0_index[i]))
        else:
            index_list.extend([i + 1] * (b0_index[i + 1] - b0_index[i]))
    index_array = np.asarray(index_list)
    try:
        len(index_list) == vols
    except ValueError:
        raise ValueError(
            "It seems that you do not define the index file for FSL eddy correctly!"
        )
    np.savetxt(out_index, index_array.T)

    return out_index


def generate_acq_file(
    in_dwi, fsl_phase_encoding_direction, total_readout_time, image_id=None
):
    """Generate [`image_id`]_acq.txt file for FSL eddy command.

    Args:
        in_dwi (str): DWI file.
        fsl_phase_encoding_direction (str): PhaseEncodingDirection from BIDS specifications in FSL format (i.e. x/y/z instead of i/j/k).
        total_readout_time (str): TotalReadoutTime from BIDS specifications.
        image_id (str, optional): Optional prefix. Defaults to None.

    Returns:
        out_acq: [`image_id`]_acq.txt or acq.txt file.
    """
    import os

    import nibabel as nb
    import numpy as np

    if image_id:
        out_acq = os.path.abspath(f"{image_id}_acq.txt")
    else:
        out_acq = os.path.abspath("acq.txt")
    vols = nb.load(in_dwi).get_data().shape[-1]
    arr = np.ones([vols, 4])
    for i in range(vols):
        if fsl_phase_encoding_direction == "y-":
            arr[i, :] = np.array((0, -1, 0, total_readout_time))
        elif fsl_phase_encoding_direction == "y":
            arr[i, :] = np.array((0, 1, 0, total_readout_time))
        elif fsl_phase_encoding_direction == "x":
            arr[i, :] = np.array((1, 0, 0, total_readout_time))
        elif fsl_phase_encoding_direction == "x-":
            arr[i, :] = np.array((-1, 0, 0, total_readout_time))
        elif fsl_phase_encoding_direction == "z":
            arr[i, :] = np.array((0, 1, 0, total_readout_time))
        elif fsl_phase_encoding_direction == "z-":
            arr[i, :] = np.array((0, 0, -1, total_readout_time))
        else:
            raise RuntimeError(
                f"FSL PhaseEncodingDirection (found value: {fsl_phase_encoding_direction}) "
                f"is unknown, it should be a value in (x, y, z, x-, y-, z-)"
            )

    np.savetxt(out_acq, arr, fmt="%d " * 3 + "%f")

    return out_acq


def bids_dir_to_fsl_dir(bids_dir):
    """Converts BIDS PhaseEncodingDirection parameters (i,j,k,i-,j-,k-) to FSL direction (x,y,z,x-,y-,z-)."""
    fsl_dir = bids_dir.lower()
    if "i" not in fsl_dir and "j" not in fsl_dir and "k" not in fsl_dir:
        raise ValueError(
            f"Unknown PhaseEncodingDirection {fsl_dir}: it should be a value in (i, j, k, i-, j-, k-)"
        )

    return fsl_dir.replace("i", "x").replace("j", "y").replace("k", "z")



from typing import Optional

import click

from clinica.pipelines import cli_param
from clinica.pipelines.cli import cli as run_cli

pipeline_name = "pet-surface"


@click.command(name=pipeline_name)
@cli_param.argument.bids_directory
@cli_param.argument.caps_directory
@cli_param.argument.acq_label
@cli_param.argument.suvr_reference_region
@cli_param.argument.pvc_psf_tsv
@cli_param.option_group.common_pipelines_options
@cli_param.option.subjects_sessions_tsv
@cli_param.option.working_directory
@cli_param.option.n_procs
@cli_param.option.yes
def cli(
    bids_directory: str,
    caps_directory: str,
    acq_label: str,
    suvr_reference_region: str,
    pvc_psf_tsv: str,
    subjects_sessions_tsv: Optional[str] = None,
    working_directory: Optional[str] = None,
    n_procs: Optional[int] = None,
    yes: bool = False,
) -> None:
    """Surface-based processing of PET images.

       ACQ_LABEL corresponds to the label given to the PET acquisition, specifying the tracer used.
    Frequently used values are '18FFDG' or '18FAV45'.

       The reference region must be specified to perform intensity normalization.
    Accepted values include: 'pons', 'cerebellumPons', 'pons2', 'cerebellumPons2'.

       PVC_PSF_TSV is the TSV file containing the psf_x, psf_y and psf_z of the PSF for each PET image.

    Prerequisite: You need to have performed the t1-freesurfer pipeline on your T1-weighted MR images.

    See https://aramislab.paris.inria.fr/clinica/docs/public/latest/Pipelines/PET_Surface/
    """
    from networkx import Graph

    from clinica.utils.ux import print_end_pipeline

    from .pet_surface_pipeline import PetSurface

    parameters = {
        "acq_label": acq_label,
        "suvr_reference_region": suvr_reference_region,
        "pvc_psf_tsv": pvc_psf_tsv,
        "longitudinal": False,
        "skip_question": yes,
    }

    pipeline = PetSurface(
        bids_directory=bids_directory,
        caps_directory=caps_directory,
        tsv_file=subjects_sessions_tsv,
        base_dir=working_directory,
        parameters=parameters,
        name=pipeline_name,
    )

    exec_pipeline = (
        pipeline.run(plugin="MultiProc", plugin_args={"n_procs": n_procs})
        if n_procs
        else pipeline.run()
    )

    if isinstance(exec_pipeline, Graph):
        print_end_pipeline(
            pipeline_name, pipeline.base_dir, pipeline.base_dir_was_specified
        )


run_cli.add_command(cli)

if __name__ == "__main__":
    cli()



# coding: utf8


from pathlib import Path

from clinica.utils.stream import cprint


def compute_atlas(
    caps_directory: str, to_process_with_atlases: tuple, path_to_atlas: str
):
    import os
    import subprocess
    from pathlib import Path
    from tempfile import mkdtemp

    from clinica.compat import errno
    from clinica.utils.stream import cprint

    subject_dir = ""
    image_id = ""
    atlas = ""

    # we iterate over left and right hemisphere
    for path in Path(path_to_atlas).rglob(
        "*" + to_process_with_atlases[0] + "_6p0.gcs"
    ):
        hemisphere = Path(path.stem).stem
        atlas_name = Path(path.stem).suffix[1:].split("_")[0]
        if "long" not in to_process_with_atlases[1]:
            sub, ses = to_process_with_atlases[1].split("_")
            substitute_dir = (
                Path(caps_directory)
                / "subjects"
                / sub
                / ses
                / ("t1/freesurfer_cross_sectional")
            )
            subject_dir = substitute_dir
            path_to_freesurfer_cross = Path(substitute_dir) / Path(
                to_process_with_atlases[1]
            )
            output_path_annot = (
                path_to_freesurfer_cross
                / "label"
                / (hemisphere + "." + atlas_name + ".annot")
            )
            sphere_reg = (
                path_to_freesurfer_cross / "surf" / (hemisphere + ".sphere.reg")
            )
            output_path_stats = (
                path_to_freesurfer_cross
                / "stats"
                / (hemisphere + "." + atlas_name + ".stats")
            )
            subjid = to_process_with_atlases[1]
        else:
            sub, ses, long = to_process_with_atlases[1].split("_")
            substitute_dir = mkdtemp()
            os.makedirs(substitute_dir, exist_ok=True)
            subjid = sub + "_" + ses + ".long." + sub + "_" + long
            subject_dir = (
                Path(caps_directory)
                / "subjects"
                / sub
                / ses
                / "t1"
                / long
                / "freesurfer_longitudinal"
            )
            path_to_freesurfer_cross = subject_dir / (
                sub + "_" + ses + ".long." + sub + "_" + long
            )
            try:
                os.symlink(
                    path_to_freesurfer_cross, os.path.join(substitute_dir, subjid)
                )
            except FileExistsError as e:
                if e.errno != errno.EEXIST:  # EEXIST: folder already exists
                    raise e
            substitute_dir_id = os.path.join(substitute_dir, subjid)

            output_path_annot = (
                Path(substitute_dir_id)
                / "label"
                / (hemisphere + "." + atlas_name + ".annot")
            )
            sphere_reg = Path(substitute_dir_id) / "surf" / (hemisphere + ".sphere.reg")
            output_path_stats = (
                Path(substitute_dir_id)
                / "stats"
                / (hemisphere + "." + atlas_name + ".stats")
            )
        if not os.path.isfile(sphere_reg):
            cprint(
                f"The {hemisphere}.sphere.reg file appears to be missing. The data for {to_process_with_atlases[1]} will not be processed with {to_process_with_atlases[0]}",
                lvl="warning",
            )
        mris_ca_label_command = f"mris_ca_label -sdir {substitute_dir} {subjid} {hemisphere} {path_to_freesurfer_cross}/surf/{hemisphere}.sphere.reg {path} {output_path_annot}"
        a = subprocess.run(mris_ca_label_command, shell=True, capture_output=True)
        mris_anatomical_stats_command = f"export SUBJECTS_DIR={substitute_dir}; mris_anatomical_stats -a {output_path_annot} -f {output_path_stats} -b {subjid} {hemisphere}"
        c = subprocess.run(
            mris_anatomical_stats_command, shell=True, capture_output=True
        )
        image_id, atlas = (
            to_process_with_atlases[1],
            to_process_with_atlases[0],
        )
    return subject_dir, image_id, atlas


def write_tsv_files(subject_dir: str, image_id: str, atlas: str) -> str:
    """
    Generate statistics TSV files in `subjects_dir`/regional_measures folder for `image_id`.

    Notes:
        We do not need to check the line "finished without error" in scripts/recon-all.log.
        If an error occurs, it will be detected by Nipype and the next nodes (including
        write_tsv_files will not be called).
    """
    import os

    from clinica.utils.freesurfer import generate_regional_measures_alt
    from clinica.utils.stream import cprint

    if "long" not in image_id:
        folder = image_id
    else:
        sub, ses, long = image_id.split("_")
        folder = sub + "_" + ses + ".long." + sub + "_" + long
    if os.path.isfile(os.path.join(subject_dir, folder, "mri", "aparc+aseg.mgz")):
        generate_regional_measures_alt(subject_dir, folder, atlas)
    else:
        cprint(
            msg=(
                f"{image_id.replace('_', ' | ')} does not contain "
                f"mri/aseg+aparc.mgz file. Creation of regional_measures/ folder will be skipped."
            ),
            lvl="warning",
        )
    return image_id



from optparse import Option
from typing import List, Optional

import click

import clinica.pydra.engine_utils as pydra_utils
import clinica.pydra.t1_volume.create_dartel.pipeline as pydra_create_dartel
from clinica.pipelines import cli_param
from clinica.pipelines.cli import cli as run_cli

pipeline_name = "pydra-create-dartel"


@click.command(name=pipeline_name, hidden=True)
@cli_param.argument.bids_directory
@cli_param.argument.caps_directory
@cli_param.argument.group_label
@cli_param.option_group.common_pipelines_options
@cli_param.option.subjects_sessions_tsv
@cli_param.option.working_directory
@cli_param.option.n_procs
@cli_param.option_group.advanced_pipeline_options
@cli_param.option.dartel_tissues
def cli(
    bids_directory: str,
    caps_directory: str,
    group_label: str,
    dartel_tissues: List[int] = (1, 2, 3),
    subjects_sessions_tsv: Optional[str] = None,
    working_directory: Optional[str] = None,
    n_procs: Optional[int] = None,
) -> None:
    """Pydra Inter-subject registration using Dartel (creating a new Dartel template).

       GROUP_LABEL is an user-defined identifier to target a specific group of subjects. For this pipeline, it is associated to the DARTEL template that you had created when running the t1-volume pipeline.

    https://aramislab.paris.inria.fr/clinica/docs/public/latest/Pipelines/T1_Volume/
    """

    parameters = {"group_label": group_label, "dartel_tissues": dartel_tissues}
    t1_volume_create_dartel_pipeline = pydra_create_dartel.t1volume_create_dartel(
        name="t1-volume-tissue-segmentation-pydra",
        input_dir=bids_directory,
        output_dir=caps_directory,
        parameters=parameters,
    )
    pydra_utils.run(t1_volume_create_dartel_pipeline)


run_cli.add_command(cli)

if __name__ == "__main__":
    cli()



from typing import List, Optional

import click

from clinica.iotools.converters import cli_param

ALL_MODALITIES = ("T1", "PET_FDG", "PET_AMYLOID", "PET_TAU", "DWI", "FLAIR", "fMRI")


@click.command(name="adni-to-bids")
@cli_param.dataset_directory
@cli_param.clinical_data_directory
@cli_param.bids_directory
@cli_param.subjects_list
@click.option(
    "-c",
    "--clinical-data-only",
    is_flag=True,
    help="Convert clinical data only.",
)
@click.option(
    "-f",
    "--force-new-extraction",
    is_flag=True,
    help="Force new data extraction.",
)
@click.option(
    "-m",
    "--modalities",
    multiple=True,
    type=click.Choice(ALL_MODALITIES),
    default=ALL_MODALITIES,
    help="Convert only the selected modality. By default, all available modalities are converted.",
)
@click.option(
    "-xml", "--xml_path", help="Path to the root directory containing the xml metadata."
)
def cli(
    dataset_directory: str,
    clinical_data_directory: str,
    bids_directory: str,
    xml_path: Optional[str] = None,
    subjects_list: Optional[str] = None,
    clinical_data_only: bool = False,
    force_new_extraction: bool = False,
    modalities: List[str] = ALL_MODALITIES,
) -> None:
    """ADNI to BIDS converter.

    Convert the imaging and clinical data of ADNI (https://adni.loni.usc.edu/), located in DATASET_DIRECTORY and
    CLINICAL_DATA_DIRECTORY respectively, to a BIDS dataset in the target BIDS_DIRECTORY.
    """
    from clinica.iotools.converters.adni_to_bids.adni_to_bids import AdniToBids
    from clinica.utils.exceptions import ClinicaParserError

    adni_to_bids = AdniToBids()
    adni_to_bids.check_adni_dependencies()

    if clinical_data_only and force_new_extraction:
        raise ClinicaParserError(
            "Arguments `clinical_data_only` and `force_new_extraction` are mutually exclusive."
        )

    if not clinical_data_only:
        adni_to_bids.convert_images(
            dataset_directory,
            clinical_data_directory,
            bids_directory,
            subjects_list,
            modalities,
            force_new_extraction,
        )

    adni_to_bids.convert_clinical_data(
        clinical_data_dir=clinical_data_directory,
        out_path=bids_directory,
        clinical_data_only=clinical_data_only,
        subjects_list_path=subjects_list,
        xml_path=xml_path,
    )


if __name__ == "__main__":
    cli()



# coding: utf-8

"""Module for converting fMRI of ADNI."""


def convert_adni_fmri(
    source_dir, csv_dir, dest_dir, conversion_dir, subjs_list=None, mod_to_update=False
):
    """Convert fMR images of ADNI into BIDS format.

    Args:
        source_dir: path to the ADNI directory
        csv_dir: path to the clinical data directory
        dest_dir: path to the destination BIDS directory
        conversion_dir: path to the TSV files including the paths to original images
        subjs_list: subjects list
        mod_to_update: If True, pre-existing images in the BIDS directory will be erased and extracted again.
    """
    from os import path

    import pandas as pd

    from clinica.iotools.converters.adni_to_bids.adni_utils import paths_to_bids
    from clinica.utils.stream import cprint

    if not subjs_list:
        adni_merge_path = path.join(csv_dir, "ADNIMERGE.csv")
        adni_merge = pd.read_csv(adni_merge_path, sep=",", low_memory=False)
        subjs_list = list(adni_merge.PTID.unique())

    cprint(
        f"Calculating paths of fMRI images. Output will be stored in {conversion_dir}."
    )
    images = compute_fmri_path(source_dir, csv_dir, subjs_list, conversion_dir)
    cprint("Paths of fMRI images found. Exporting images into BIDS ...")
    # fmri_paths_to_bids(dest_dir, images)
    paths_to_bids(images, dest_dir, "fmri", mod_to_update=mod_to_update)
    cprint(msg="fMRI conversion done.", lvl="debug")


def compute_fmri_path(source_dir, csv_dir, subjs_list, conversion_dir):
    """Compute the paths to fMR images.

    Args:
        source_dir: path to the ADNI directory
        csv_dir: path to the clinical data directory
        subjs_list: subjects list
        conversion_dir: path to the TSV files including the paths to original images

    Returns:
        pandas Dataframe containing the path for each fmri
    """
    from os import path

    import pandas as pd

    from clinica.iotools.converters.adni_to_bids.adni_utils import (
        find_image_path,
        visits_to_timepoints,
    )

    fmri_col = [
        "Subject_ID",
        "VISCODE",
        "Visit",
        "Sequence",
        "Scan_Date",
        "Study_ID",
        "Field_Strength",
        "Series_ID",
        "Image_ID",
    ]
    fmri_df = pd.DataFrame(columns=fmri_col)
    fmri_dfs_list = []

    # Loading needed .csv files
    adni_merge = pd.read_csv(
        path.join(csv_dir, "ADNIMERGE.csv"), sep=",", low_memory=False
    )

    mayo_mri_qc = pd.read_csv(
        path.join(csv_dir, "MAYOADIRL_MRI_IMAGEQC_12_08_15.csv"),
        sep=",",
        low_memory=False,
    )
    mayo_mri_qc = mayo_mri_qc[mayo_mri_qc.series_type == "fMRI"]
    mayo_mri_qc.columns = [x.upper() for x in mayo_mri_qc.columns]

    mayo_mri_qc3 = pd.read_csv(
        path.join(csv_dir, "MAYOADIRL_MRI_QUALITY_ADNI3.csv"), sep=",", low_memory=False
    )
    mayo_mri_qc3 = mayo_mri_qc3[mayo_mri_qc3.SERIES_TYPE == "EPB"]

    # Concatenating visits in both QC files
    mayo_mri_qc = pd.concat(
        [mayo_mri_qc, mayo_mri_qc3], axis=0, ignore_index=True, sort=False
    )

    mri_list = pd.read_csv(path.join(csv_dir, "MRILIST.csv"), sep=",", low_memory=False)

    # Selecting only fMRI images that are not Multiband
    mri_list = mri_list[
        mri_list.SEQUENCE.str.contains("MRI")
    ]  # 'MRI' includes all fMRI and fMRI scans, but not others
    unwanted_sequences = ["MB"]
    mri_list = mri_list[
        mri_list.SEQUENCE.map(
            lambda x: not any(subs in x for subs in unwanted_sequences)
        )
    ]

    # We will convert the images for each subject in the subject list
    for subj in subjs_list:

        # Filter ADNIMERGE, MRI_LIST and QC for only one subject and sort the rows/visits by examination date
        adnimerge_subj = adni_merge[adni_merge.PTID == subj]
        adnimerge_subj = adnimerge_subj.sort_values("EXAMDATE")

        mri_list_subj = mri_list[mri_list.SUBJECT == subj]
        mri_list_subj = mri_list_subj.sort_values("SCANDATE")

        mayo_mri_qc_subj = mayo_mri_qc[mayo_mri_qc.RID == int(subj[-4:])]

        # Obtain corresponding timepoints for the subject visits
        visits = visits_to_timepoints(subj, mri_list_subj, adnimerge_subj, "fMRI")

        for visit_info in visits.keys():
            timepoint = visit_info[0]
            visit_str = visits[visit_info]

            visit_mri_list = mri_list_subj[mri_list_subj.VISIT == visit_str]
            image = fmri_image(
                subj, timepoint, visits[visit_info], visit_mri_list, mayo_mri_qc_subj
            )

            if image is not None:
                row_to_append = pd.DataFrame(
                    image,
                    index=[
                        "i",
                    ],
                )
                fmri_dfs_list.append(row_to_append)

    if fmri_dfs_list:
        fmri_df = pd.concat(fmri_dfs_list, ignore_index=True)

    # Exceptions
    # ==========
    conversion_errors = [
        ("006_S_4485", "m84"),
        ("123_S_4127", "m96"),
        # Eq_1
        ("094_S_4503", "m24"),
        ("009_S_4388", "m72"),
        ("036_S_6088", "bl"),
        ("036_S_6134", "bl"),
        ("016_S_6802", "bl"),
        ("016_S_6816", "bl"),
        ("126_S_4891", "m84"),
        # Multiple images
        ("029_S_2395", "m72"),
    ]

    # Removing known exceptions from images to convert
    if not fmri_df.empty:
        error_ind = fmri_df.index[
            fmri_df.apply(
                lambda x: ((x.Subject_ID, x.VISCODE) in conversion_errors), axis=1
            )
        ]
        fmri_df.drop(error_ind, inplace=True)

    # Checking for images paths in filesystem
    images = find_image_path(fmri_df, source_dir, "fMRI", "S", "Series_ID")
    images.to_csv(path.join(conversion_dir, "fmri_paths.tsv"), sep="\t", index=False)

    return images


def fmri_image(subject_id, timepoint, visit_str, visit_mri_list, mri_qc_subj):
    """
    One image among those in the input list is chosen according to QC
    and then correspoding metadata is extracted to a dictionary.

    Args:
        subject_id: Subject identifier
        timepoint: Visit code
        visit_str: Visit name
        visit_mri_list: List of images metadata
        mri_qc_subj: Dataframe containing list of QC of scans for the subject

    Returns: dictionary - contains image metadata
    """
    from clinica.iotools.converters.adni_to_bids.adni_utils import (
        replace_sequence_chars,
        select_image_qc,
    )

    mri_qc_subj.columns = [x.lower() for x in mri_qc_subj.columns]
    sel_image = select_image_qc(list(visit_mri_list.IMAGEUID), mri_qc_subj)
    if not sel_image:
        return None

    sel_scan = visit_mri_list[visit_mri_list.IMAGEUID == sel_image].iloc[0]

    image_dict = {
        "Subject_ID": subject_id,
        "VISCODE": timepoint,
        "Visit": visit_str,
        "Sequence": replace_sequence_chars(sel_scan.SEQUENCE),
        "Scan_Date": sel_scan["SCANDATE"],
        "Study_ID": str(int(sel_scan.STUDYID)),
        "Series_ID": str(int(sel_scan.SERIESID)),
        "Image_ID": str(int(sel_scan.IMAGEUID)),
        "Field_Strength": sel_scan.MAGSTRENGTH,
    }

    return image_dict



from typing import List, Optional, Tuple


def _zip_unzip_nii(in_file: str, same_dir: bool, compress: bool):
    import gzip
    import operator
    import shutil
    from os import getcwd
    from os.path import abspath, join
    from pathlib import Path

    from nipype.utils.filemanip import split_filename
    from traits.trait_base import _Undefined

    if (in_file is None) or isinstance(in_file, _Undefined):
        return None

    if isinstance(in_file, list):
        return [_zip_unzip_nii(f, same_dir, compress) for f in in_file]

    in_file = Path(in_file)

    op = operator.eq if compress else operator.ne
    if op(in_file.suffix, ".gz"):
        return str(in_file)

    if not in_file.exists():
        raise FileNotFoundError(f"File {in_file} does not exist.")

    orig_dir, base, ext = split_filename(str(in_file))
    new_ext = ext + ".gz" if compress else ext[:-3]
    out_file = abspath(join(orig_dir if same_dir else getcwd(), base + new_ext))

    outer = open if compress else gzip.open
    inner = gzip.open if compress else open
    with outer(in_file, "rb") as f_in:
        with inner(out_file, "wb") as f_out:
            shutil.copyfileobj(f_in, f_out)

    return out_file


def zip_nii(in_file: str, same_dir: bool = False) -> str:
    """Compress the provided file(s).

    .. note::
        If a list of file paths is provided, then this function
        will be called on each item and a list of paths to the
        resulting zip files will be returned.
        The type hint is not reflecting this fact because of the
        self-contained requirement from Nipype which is preventing
        the usage of types outside of Python's builtins.

    Parameters
    ----------
    in_file: str
        Path to file to be zipped as a string.

    same_dir: bool, optional
        If True, the zip file is written in the same directory as the
        provided file. Otherwise, it is written in the current working
        directory. Default=False.

    Returns
    -------
    out_file: str
        Path to the resulting zip file.

    Notes
    -----
    If the provided file is already zipped (that is, its extension
    is '.gz'), then nothing is done and the path to the provided file
    is returned.

    Raises
    ------
    FileNotFoundError
        If the provided file does not exist.
    """
    from clinica.utils.filemanip import _zip_unzip_nii  # noqa

    return _zip_unzip_nii(in_file, same_dir, compress=True)


def unzip_nii(
    in_file: str,
    same_dir: bool = False,
) -> str:
    """Decompress the provided file(s).

    .. note::
        If a list of file paths is provided, then this function
        will be called on each item and a list of paths to the
        resulting unzipped files will be returned.
        The type hint is not reflecting this fact because of the
        self-contained requirement from Nipype which is preventing
        the usage of types outside of Python's builtins.

    Parameters
    ----------
    in_file: str
        Path to file to be zipped.

    same_dir: bool, optional
        If True, the unzipped file is written in the same directory as the
        provided zip file. Otherwise, it is written in the current working
        directory. Default=False.

    Returns
    -------
    out_file: str
        Path to the resulting unzipped file.

    Notes
    -----
    If the provided file is not zipped (that is, its extension
    is different from '.gz'), then nothing is done and the path
    to the provided file is returned.

    Raises
    ------
    FileNotFoundError
        If the provided file does not exist.
    """
    from clinica.utils.filemanip import _zip_unzip_nii  # noqa

    return _zip_unzip_nii(in_file, same_dir, compress=False)


def save_participants_sessions(
    participant_ids: List[str],
    session_ids: List[str],
    out_folder: str,
    out_file: Optional[str] = "participants.tsv",
) -> None:
    """Save the participants and sessions to TSV.

    Parameters
    ----------
    participant_ids: List[str]
        List of participant IDs.

    session_ids: List[str]
        List of session IDs.

    out_folder: str
        Folder where the resulting TSV file should be written.

    out_file: str, optional
        Name of the TSV file. Default="participants.tsv".

    Raises
    ------
    ValueError
        If provided `participant_ids` and `session_ids` do not have
        the same length.
    """
    import os
    from pathlib import Path

    import pandas as pd

    from clinica.utils.stream import cprint

    if len(participant_ids) != len(session_ids):
        raise ValueError(
            "The number of participant IDs is not equal to the number of session IDs."
        )

    out_folder = Path(out_folder)
    out_folder.mkdir(parents=True, exist_ok=True)
    tsv_file = out_folder / out_file

    data = pd.DataFrame(
        {
            "participant_id": participant_ids,
            "session_id": session_ids,
        }
    )
    try:
        data.to_csv(tsv_file, sep="\t", index=False, encoding="utf-8")
    except Exception as e:
        cprint(msg=f"Impossible to save {out_file} with pandas", lvl="error")
        raise e


def _check_bids_or_caps_compliance(filename: str, sep: str):
    import re

    m = re.search(sep.join([r"(sub-[a-zA-Z0-9]+)", r"(ses-[a-zA-Z0-9]+)"]), filename)
    if not m:
        raise ValueError(
            f"Input filename {filename} is not in a BIDS or CAPS compliant format."
            " It does not contain the subject and session information."
        )

    return m


def get_subject_id(bids_or_caps_file: str) -> str:
    """Extract the subject ID from a BIDS or CAPS file path.

    The subject ID is defined as

        sub-<participant_id>_ses-<session_label>

    In other words, it is the concatenation of the subject and session labels,
    separated by "_".

    Parameters
    ----------
    bids_or_caps_file: str
        Path to a file from a BIDS or CAPS folder.

    Returns
    -------
    subject_id: str
        The subject ID corresponding to the given file.

    Examples
    --------
    >>> get_subject_id("sub-01/ses-M000/pet/sub-01_ses-M000_trc-18FAV45_pet.nii.gz")
    'sub-01_ses-M000'
    >>> get_subject_id("sub-01_ses-M000_trc-18FAV45_pet.nii.gz")
    Traceback (most recent call last):
    ValueError: Input filename sub-01_ses-M000_trc-18FAV45_pet.nii.gz is not in a BIDS or CAPS compliant format. It does not contain the subject and session information.  # noqa

    See also
    --------
    extract_image_ids
    """
    match = _check_bids_or_caps_compliance(bids_or_caps_file, sep="/")
    subject_id = match.group(1) + "_" + match.group(2)

    return subject_id


def get_filename_no_ext(filename: str) -> str:
    """Get the filename without the extension.

    Parameters
    ----------
    filename: str
        The full filename from which to extract the extension out.

    Returns
    -------
    filename_no_ext: str
        The filename with extension removed.

    Examples
    --------
    >>> get_filename_no_ext("foo.nii.gz")
    'foo'
    >>> get_filename_no_ext("sub-01/ses-M000/sub-01_ses-M000.tar.gz")
    'sub-01_ses-M000'
    """
    from pathlib import PurePath

    stem = PurePath(filename).stem
    while "." in stem:
        stem = PurePath(stem).stem

    return stem


def extract_image_ids(bids_or_caps_files: List[str]) -> List[str]:
    """Extract the image IDs from a list of BIDS or CAPS files.

    .. warning::
        The image ID is the same as the subject ID from function
        `get_subject_id()` but is extracted from the filename rather
        than from the path. (See examples section).

    Parameters
    ----------
    bids_or_caps_files: List[str]
        List of file paths from which to extract the image IDs.

    Returns
    -------
    id_bids_or_caps_files: List[str]
        List of extracted image IDs.

    Examples
    --------
    >>> extract_image_ids(["sub-01/ses-M000/pet/sub-01_ses-M000_trc-18FAV45_pet.nii.gz"])
    ['sub-01_ses-M000']

    Beware, the image IDs is extracted from the filename and not from the folder names
    as it is the case for `get_subject_id()`:

    >>> extract_image_ids(["foo/bar/baz/sub-foo/ses-bar/foooo/sub-01_ses-M000_foo.json"])
    ['sub-01_ses-M000']

    See also
    --------
    get_subject_id
    """
    id_bids_or_caps_files = []
    for f in bids_or_caps_files:
        match = _check_bids_or_caps_compliance(f, sep="_")
        id_bids_or_caps_files.append(match.group())

    return id_bids_or_caps_files


def extract_subjects_sessions_from_filename(
    bids_or_caps_files: List[str],
) -> Tuple[List[str], List[str]]:
    """Extract the subject and session labels from a list of BIDS or CAPS files.

    Parameters
    ----------
    bids_or_caps_files: List[str]
        List of file paths for which to extract the subject and session labels.

    Returns
    -------
    subject_ids: List[str]
        List of subject labels in the same order as `bids_or_caps_files`.

    session_ids: List[str]
        List of session labels in the same order as `bids_or_caps_files`.

    Examples
    --------
    >>> extract_subjects_sessions_from_filename([
    ...     "sub-01/ses-M000/pet/sub-01_ses-M000_trc-18FAV45_pet.nii.gz",
    ...     "foo/bar/baz/sub-foo/ses-bar/foooo/sub-01_ses-M000_foo.json", "sub-01_ses-M000.tar.gz",
    ...])
    (['sub-01', 'sub-01', 'sub-01'], ['ses-M000', 'ses-M000', 'ses-M000'])

    See also
    --------
    extract_image_ids
    """
    id_bids_or_caps_files = extract_image_ids(bids_or_caps_files)
    split = [image_id.split("_") for image_id in id_bids_or_caps_files]
    subject_ids = [p_id[0] for p_id in split]
    session_ids = [s_id[1] for s_id in split]
    return subject_ids, session_ids


def extract_crash_files_from_log_file(filename: str) -> List[str]:
    """Extract crash files (*.pklz) from `filename`.

    Parameters
    ----------
    filename: str
        Path to log file for which the crash files should be extracted.

    Returns
    -------
    crash_files: List[str]
        List of crash files.
    """
    import re
    from pathlib import Path

    filename = Path(filename)
    if not filename.is_file():
        raise ValueError(
            f"extract_crash_files_from_log_file: filename parameter is not a file ({filename})"
        )
    crash_files = []
    with open(filename, "r") as log_file:
        for line in log_file:
            if re.match("(.*)crashfile:(.*)", line):
                crash_files.append(line.replace("\t crashfile:", "").replace("\n", ""))

    return crash_files


def read_participant_tsv(tsv_file: str) -> Tuple[List[str], List[str]]:
    """Extract participant IDs and session IDs from TSV file.

    Parameters
    ----------
    tsv_file: str
        Participant TSV file from which to extract the participant and session IDs.

    Returns
    -------
    participants: List[str]
        List of participant IDs.

    sessions: List[str]
        List of session IDs.

    Raises
    ------
    ClinicaException
        If `tsv_file` is not a file.
        If `participant_id` or `session_id` column is missing from TSV file.

    Examples
    --------
    >>> dframe = pd.DataFrame({
    ...     "participant_id": ["sub-01", "sub-01", "sub-02"],
    ...     "session_id": ["ses-M000", "ses-M006", "ses-M000"],
    ...})
    >>> dframe.to_csv("participants.tsv", sep="\t")
    >>> read_participant_tsv("participant.tsv")
    (["sub-01", "sub-01", "sub-02"], ["ses-M000", "ses-M006", "ses-M000"])
    """
    import pandas as pd

    from clinica.utils.exceptions import ClinicaException

    try:
        df = pd.read_csv(tsv_file, sep="\t")
    except FileNotFoundError:
        raise ClinicaException(
            "The TSV file you gave is not a file.\nError explanations:\n"
            f"\t- Clinica expected the following path to be a file: {tsv_file}\n"
            "\t- If you gave relative path, did you run Clinica on the good folder?"
        )

    for column in ("participant_id", "session_id"):
        if column not in list(df.columns.values):
            raise ClinicaException(
                f"The TSV file does not contain {column} column (path: {tsv_file})"
            )

    return (
        [sub.strip(" ") for sub in list(df.participant_id)],
        [ses.strip(" ") for ses in list(df.session_id)],
    )


def extract_metadata_from_json(json_file: str, list_keys: List[str]) -> List[str]:
    """Extract fields from JSON file.

    Parameters
    ----------
    json_file: str
        Path to JSON file from which to extract metadata.

    list_keys: List[str]
        List of keys to extract.

    Returns
    -------
    list_values: List[str]
        List of values extracted from the JSON file and corresponding to the keys.
    """
    import json

    from clinica.utils.exceptions import ClinicaException

    try:
        with open(json_file, "r") as file:
            data = json.load(file)
    except FileNotFoundError:
        raise FileNotFoundError(
            f"Clinica could not open the following JSON file: {json_file}"
        )
    list_values = []
    missing_keys = []
    for key in list_keys:
        if key in data:
            list_values.append(data[key])
        else:
            missing_keys.append(key)
    if len(missing_keys) > 0:
        error_message = f"Clinica could not find the following keys in the following JSON file: {json_file}:\n- "
        error_message += "\n- ".join(missing_keys)
        raise ClinicaException(error_message)

    return list_values



from nipype import config

import clinica.pipelines.engine as cpe

# Use hash instead of parameters for iterables folder names
# Otherwise path will be too long and generate OSError
cfg = dict(execution={"parameterize_dirs": False})
config.update_config(cfg)


class DwiPreprocessingUsingPhaseDiffFMap(cpe.Pipeline):
    """DWI Preprocessing using phase difference fieldmap.

    Ideas for improvement:
        - Use promising sdcflows workflows and/or dMRIprep

    Note:
        Some reading regarding the reproducibility of FSL eddy command:
        https://www.jiscmail.ac.uk/cgi-bin/webadmin?A2=fsl;1ccf038f.1608

    Returns:
        A clinica pipeline object containing the DwiPreprocessingUsingPhaseDiffFMap pipeline.
    """

    @staticmethod
    def get_processed_images(caps_directory, subjects, sessions):
        import os

        from clinica.utils.filemanip import extract_image_ids
        from clinica.utils.input_files import DWI_PREPROC_NII
        from clinica.utils.inputs import clinica_file_reader

        image_ids = []
        if os.path.isdir(caps_directory):
            preproc_files, _ = clinica_file_reader(
                subjects, sessions, caps_directory, DWI_PREPROC_NII, False
            )
            image_ids = extract_image_ids(preproc_files)
        return image_ids

    def check_pipeline_parameters(self):
        """Check pipeline parameters."""
        from clinica.utils.stream import cprint

        self.parameters.setdefault("low_bval", 5)
        low_bval = self.parameters["low_bval"]
        if low_bval < 0:
            raise ValueError(
                f"The low_bval is negative ({low_bval}): it should be zero or close to zero."
            )
        if self.parameters["low_bval"] > 100:
            cprint(
                f"The low_bval parameter is {low_bval}: it should be close to zero.",
                lvl="warning",
            )

        self.parameters.setdefault("use_cuda", False)
        self.parameters.setdefault("initrand", False)

    def check_custom_dependencies(self):
        """Check dependencies that can not be listed in the `info.json` file."""
        from clinica.utils.check_dependency import is_binary_present
        from clinica.utils.exceptions import ClinicaMissingDependencyError

        if self.parameters["use_cuda"]:
            if not is_binary_present("eddy_cuda"):
                raise ClinicaMissingDependencyError(
                    "[Error] FSL eddy with CUDA was set but Clinica could not find eddy_cuda in your PATH environment. "
                    "Check that  https://fsl.fmrib.ox.ac.uk/fsl/fslwiki/eddy/UsersGuide#The_eddy_executables is correctly set."
                )

    def get_input_fields(self):
        """Specify the list of possible inputs of this pipeline.

        Returns:
            List[str]: The list of inputs for the DwiPreprocessingUsingPhaseDiffFMap pipeline namely:
                * dwi: Path of the diffusion weighted image in BIDS format
                * bvec: Path of the bvec file in BIDS format
                * bval: Path of the bval file in BIDS format
                * dwi_json: Path of the DWI JSON file in BIDS format and containing
                    TotalReadoutTime and PhaseEncodingDirection metadata (see BIDS specifications)
                * fmap_magnitude: Path of the (1st) magnitude image in BIDS format
                * fmap_phasediff: Path of the phase difference image in BIDS format
                * fmap_phasediff_json: Path of the phase difference JSON file in BIDS format
                    and containing EchoTime1 & EchoTime2 metadata (see BIDS specifications)
        """
        return [
            "dwi",
            "bvec",
            "bval",
            "dwi_json",
            "fmap_magnitude",
            "fmap_phasediff",
            "fmap_phasediff_json",
        ]

    def get_output_fields(self):
        """Specify the list of possible outputs of this pipeline.

        Returns:
            List[str]: The list of outputs generated by the DwiPreprocessingUsingPhaseDiffFMap pipeline namely:
                * preproc_dwi: Path of the preprocessed DWI
                * preproc_bvec: Path of the preprocessed bvec
                * preproc_bval: Path of the preprocessed bval
                * b0_mask: Path of the b0 brainmask
                * magnitude_on_b0: Path of the smoothed calibrated fmap on b0 space
                * calibrated_fmap_on_b0: Path of the calibrated fmap on b0 space
                * smoothed_fmap_on_b0: Path of the magnitude fmap on b0 space
        """
        return [
            "preproc_dwi",
            "preproc_bvec",
            "preproc_bval",
            "b0_mask",
            "magnitude_on_b0",
            "calibrated_fmap_on_b0",
            "smoothed_fmap_on_b0",
        ]

    def build_input_node(self):
        """Build and connect an input node to the pipeline."""
        import os

        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from clinica.utils.filemanip import save_participants_sessions
        from clinica.utils.input_files import (
            DWI_BVAL,
            DWI_BVEC,
            DWI_JSON,
            DWI_NII,
            FMAP_MAGNITUDE1_NII,
            FMAP_PHASEDIFF_JSON,
            FMAP_PHASEDIFF_NII,
        )
        from clinica.utils.inputs import clinica_list_of_files_reader
        from clinica.utils.stream import cprint
        from clinica.utils.ux import print_images_to_process

        list_bids_files = clinica_list_of_files_reader(
            self.subjects,
            self.sessions,
            self.bids_directory,
            [
                DWI_NII,
                DWI_BVEC,
                DWI_BVAL,
                DWI_JSON,
                FMAP_MAGNITUDE1_NII,
                FMAP_PHASEDIFF_NII,
                FMAP_PHASEDIFF_JSON,
            ],
            raise_exception=True,
        )

        # Save subjects to process in <WD>/<Pipeline.name>/participants.tsv
        folder_participants_tsv = os.path.join(self.base_dir, self.name)
        save_participants_sessions(
            self.subjects, self.sessions, folder_participants_tsv
        )

        if len(self.subjects):
            print_images_to_process(self.subjects, self.sessions)
            cprint(
                f"List available in {os.path.join(folder_participants_tsv, 'participants.tsv')}"
            )
            cprint(
                "Computational time will depend of the number of volumes in your DWI dataset and the use of CUDA."
            )

        read_node = npe.Node(
            name="ReadingFiles",
            iterables=[
                ("dwi", list_bids_files[0]),
                ("bvec", list_bids_files[1]),
                ("bval", list_bids_files[2]),
                ("dwi_json", list_bids_files[3]),
                ("fmap_magnitude", list_bids_files[4]),
                ("fmap_phasediff", list_bids_files[5]),
                ("fmap_phasediff_json", list_bids_files[6]),
            ],
            synchronize=True,
            interface=nutil.IdentityInterface(fields=self.get_input_fields()),
        )
        # fmt: off
        self.connect(
            [
                (read_node, self.input_node, [("dwi", "dwi"),
                                              ("bvec", "bvec"),
                                              ("bval", "bval"),
                                              ("dwi_json", "dwi_json"),
                                              ("fmap_magnitude", "fmap_magnitude"),
                                              ("fmap_phasediff", "fmap_phasediff"),
                                              ("fmap_phasediff_json", "fmap_phasediff_json")]),
            ]
        )
        # fmt: on

    def build_output_node(self):
        """Build and connect an output node to the pipeline."""
        import nipype.interfaces.io as nio
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from clinica.utils.nipype import container_from_filename, fix_join

        from .dwi_preprocessing_using_phasediff_fmap_utils import rename_into_caps

        # Find container path from DWI filename
        # =====================================
        container_path = npe.Node(
            nutil.Function(
                input_names=["bids_or_caps_filename"],
                output_names=["container"],
                function=container_from_filename,
            ),
            name="container_path",
        )

        rename_into_caps = npe.Node(
            nutil.Function(
                input_names=[
                    "in_bids_dwi",
                    "fname_dwi",
                    "fname_bval",
                    "fname_bvec",
                    "fname_brainmask",
                    "fname_magnitude",
                    "fname_fmap",
                    "fname_smoothed_fmap",
                ],
                output_names=[
                    "out_caps_dwi",
                    "out_caps_bval",
                    "out_caps_bvec",
                    "out_caps_brainmask",
                    "out_caps_magnitude",
                    "out_caps_fmap",
                    "out_caps_smoothed_fmap",
                ],
                function=rename_into_caps,
            ),
            name="rename_into_caps",
        )

        # Writing results into CAPS
        # =========================
        write_results = npe.Node(name="write_results", interface=nio.DataSink())
        write_results.inputs.base_directory = self.caps_directory
        write_results.inputs.parameterization = False

        # fmt: off
        self.connect(
            [
                (self.input_node, container_path, [("dwi", "bids_or_caps_filename")]),
                (self.input_node, rename_into_caps, [("dwi", "in_bids_dwi")]),
                (self.output_node, rename_into_caps, [("preproc_dwi", "fname_dwi"),
                                                      ("preproc_bval", "fname_bval"),
                                                      ("preproc_bvec", "fname_bvec"),
                                                      ("b0_mask", "fname_brainmask"),
                                                      ("magnitude_on_b0", "fname_magnitude"),
                                                      ("calibrated_fmap_on_b0", "fname_fmap"),
                                                      ("smoothed_fmap_on_b0", "fname_smoothed_fmap")]),
                (container_path, write_results, [(("container", fix_join, "dwi"), "container")]),
                (rename_into_caps, write_results, [("out_caps_dwi", "preprocessing.@preproc_dwi"),
                                                   ("out_caps_bval", "preprocessing.@preproc_bval"),
                                                   ("out_caps_bvec", "preprocessing.@preproc_bvec"),
                                                   ("out_caps_brainmask", "preprocessing.@b0_mask"),
                                                   ("out_caps_magnitude", "preprocessing.@magnitude"),
                                                   ("out_caps_fmap", "preprocessing.@fmap"),
                                                   ("out_caps_smoothed_fmap", "preprocessing.@smoothed_fmap")]),
            ]
        )
        # fmt: on

    def build_core_nodes(self):
        """Build and connect the core nodes of the pipeline."""
        import nipype.interfaces.ants as ants
        import nipype.interfaces.fsl as fsl
        import nipype.interfaces.mrtrix3 as mrtrix3
        import nipype.interfaces.utility as nutil
        import nipype.interfaces.utility as niu
        import nipype.pipeline.engine as npe
        from nipype.interfaces.fsl.epi import Eddy

        from clinica.utils.dwi import (
            compute_average_b0,
            generate_acq_file,
            generate_index_file,
        )

        from .dwi_preprocessing_using_phasediff_fmap_utils import (
            get_grad_fsl,
            init_input_node,
            print_end_pipeline,
        )
        from .dwi_preprocessing_using_phasediff_fmap_workflows import (
            prepare_phasediff_fmap,
        )

        # Step 0: Initialization
        # ======================
        # Initialize input parameters and print begin message
        init_node = npe.Node(
            interface=nutil.Function(
                input_names=self.get_input_fields(),
                output_names=[
                    "image_id",
                    "dwi",
                    "bvec",
                    "bval",
                    "total_readout_time",
                    "phase_encoding_direction",
                    "fmap_magnitude",
                    "fmap_phasediff",
                    "delta_echo_time",
                ],
                function=init_input_node,
            ),
            name="0-InitNode",
        )

        # Generate (bvec, bval) tuple for MRtrix interfaces
        get_grad_fsl = npe.Node(
            nutil.Function(
                input_names=["bval", "bvec"],
                output_names=["grad_fsl"],
                function=get_grad_fsl,
            ),
            name="0-GetFslGrad",
        )

        # Generate <image_id>_acq.txt for eddy
        gen_acq_txt = npe.Node(
            nutil.Function(
                input_names=[
                    "in_dwi",
                    "fsl_phase_encoding_direction",
                    "total_readout_time",
                    "image_id",
                ],
                output_names=["out_acq"],
                function=generate_acq_file,
            ),
            name="0-GenerateAcqFile",
        )

        # Generate <image_id>_index.txt for eddy
        gen_index_txt = npe.Node(
            nutil.Function(
                input_names=["in_bval", "low_bval", "image_id"],
                output_names=["out_index"],
                function=generate_index_file,
            ),
            name="0-GenerateIndexFile",
        )
        gen_index_txt.inputs.low_bval = self.parameters["low_bval"]

        # Step 1: Computation of the reference b0 (i.e. average b0 but with EPI distortions)
        # =======================================
        # Compute whole brain mask
        pre_mask_b0 = npe.Node(mrtrix3.BrainMask(), name="1a-PreMaskB0")
        pre_mask_b0.inputs.out_file = (
            "brainmask.nii.gz"  # On default, .mif file is generated
        )

        # Run eddy without calibrated fmap
        pre_eddy = npe.Node(name="1b-PreEddy", interface=Eddy())
        pre_eddy.inputs.repol = True
        pre_eddy.inputs.use_cuda = self.parameters["use_cuda"]
        pre_eddy.inputs.initrand = self.parameters["initrand"]

        # Compute the reference b0
        compute_ref_b0 = npe.Node(
            niu.Function(
                input_names=["in_dwi", "in_bval"],
                output_names=["out_b0_average"],
                function=compute_average_b0,
            ),
            name="1c-ComputeReferenceB0",
        )
        compute_ref_b0.inputs.low_bval = self.parameters["low_bval"]

        # Compute brain mask from reference b0
        mask_ref_b0 = npe.Node(
            fsl.BET(mask=True, robust=True), name="1d-MaskReferenceB0"
        )

        # Step 2: Calibrate and register FMap
        # ===================================
        # Bias field correction of the magnitude image
        bias_mag_fmap = npe.Node(
            ants.N4BiasFieldCorrection(dimension=3), name="2a-N4MagnitudeFmap"
        )
        # Brain extraction of the magnitude image
        bet_mag_fmap = npe.Node(
            fsl.BET(frac=0.4, mask=True), name="2b-BetN4MagnitudeFmap"
        )

        # Calibrate FMap
        calibrate_fmap = prepare_phasediff_fmap(name="2c-CalibrateFMap")

        # Register the BET magnitude fmap onto the BET b0
        bet_mag_fmap2b0 = npe.Node(
            interface=fsl.FLIRT(), name="2d-RegistrationBetMagToB0"
        )
        bet_mag_fmap2b0.inputs.dof = 6
        bet_mag_fmap2b0.inputs.output_type = "NIFTI_GZ"

        # Apply the transformation on the calibrated fmap
        fmap2b0 = npe.Node(interface=fsl.ApplyXFM(), name="2e-1-FMapToB0")
        fmap2b0.inputs.output_type = "NIFTI_GZ"

        # Apply the transformation on the magnitude image
        mag_fmap2b0 = fmap2b0.clone("2e-2-MagFMapToB0")

        # Smooth the registered (calibrated) fmap
        smoothing = npe.Node(interface=fsl.maths.IsotropicSmooth(), name="2f-Smoothing")
        smoothing.inputs.sigma = 4.0

        # Step 3: Run FSL eddy
        # ====================
        eddy = pre_eddy.clone("3-Eddy")

        # Step 4: Bias correction
        # =======================
        # Use implementation detailed in (Jeurissen et al., 2014)
        bias = npe.Node(mrtrix3.DWIBiasCorrect(use_ants=True), name="4-RemoveBias")

        # Step 5: Final brainmask
        # =======================
        # Compute average b0 on corrected dataset (for brain mask extraction)
        compute_avg_b0 = compute_ref_b0.clone("5a-ComputeB0Average")

        # Compute b0 mask on corrected avg b0
        mask_avg_b0 = mask_ref_b0.clone("5b-MaskB0")

        # Print end message
        print_end_message = npe.Node(
            interface=nutil.Function(
                input_names=["image_id", "final_file"], function=print_end_pipeline
            ),
            name="99-WriteEndMessage",
        )

        # Connection
        # ==========
        # fmt: off
        self.connect(
            [
                # Step 0: Initialization
                # ======================
                # Initialize input parameters and print begin message
                (self.input_node, init_node, [("dwi", "dwi"),
                                              ("bvec", "bvec"),
                                              ("bval", "bval"),
                                              ("dwi_json", "dwi_json"),
                                              ("fmap_magnitude", "fmap_magnitude"),
                                              ("fmap_phasediff", "fmap_phasediff"),
                                              ("fmap_phasediff_json", "fmap_phasediff_json")]),
                # Generate (bvec, bval) tuple for MRtrix interfaces
                (init_node, get_grad_fsl, [("bval", "bval"),
                                           ("bvec", "bvec")]),
                # Generate <image_id>_acq.txt for eddy
                (init_node, gen_acq_txt, [("dwi", "in_dwi"),
                                          ("total_readout_time", "total_readout_time"),
                                          ("phase_encoding_direction", "fsl_phase_encoding_direction"),
                                          ("image_id", "image_id")]),
                # Generate <image_id>_index.txt for eddy
                (init_node, gen_index_txt, [("bval", "in_bval"),
                                            ("image_id", "image_id")]),

                # Step 1: Computation of the reference b0 (i.e. average b0 but with EPI distortions)
                # =======================================
                # Compute whole brain mask
                (get_grad_fsl, pre_mask_b0, [("grad_fsl", "grad_fsl")]),
                (init_node, pre_mask_b0, [("dwi", "in_file")]),
                # Run eddy without calibrated fmap
                (init_node, pre_eddy, [("dwi", "in_file"),
                                       ("bval", "in_bval"),
                                       ("bvec", "in_bvec"),
                                       ("image_id", "out_base")]),
                (gen_acq_txt, pre_eddy, [("out_acq", "in_acqp")]),
                (gen_index_txt, pre_eddy, [("out_index", "in_index")]),
                (pre_mask_b0, pre_eddy, [("out_file", "in_mask")]),
                # Compute the reference b0
                (init_node, compute_ref_b0, [("bval", "in_bval")]),
                (pre_eddy, compute_ref_b0, [("out_corrected", "in_dwi")]),
                # Compute brain mask from reference b0
                (compute_ref_b0, mask_ref_b0, [("out_b0_average", "in_file")]),

                # Step 2: Calibrate and register FMap
                # ===================================
                # Bias field correction of the magnitude image
                (init_node, bias_mag_fmap, [("fmap_magnitude", "input_image")]),
                # Brain extraction of the magnitude image
                (bias_mag_fmap, bet_mag_fmap, [("output_image", "in_file")]),
                # Calibration of the FMap
                (bet_mag_fmap, calibrate_fmap, [("mask_file", "input_node.fmap_mask"),
                                                ("out_file", "input_node.fmap_magnitude")]),
                (init_node, calibrate_fmap, [("fmap_phasediff", "input_node.fmap_phasediff"),
                                             ("delta_echo_time", "input_node.delta_echo_time")]),
                # Register the BET magnitude fmap onto the BET b0
                (bet_mag_fmap, bet_mag_fmap2b0, [("out_file", "in_file")]),
                (mask_ref_b0, bet_mag_fmap2b0, [("out_file", "reference")]),
                # Apply the transformation on the magnitude image
                (bet_mag_fmap2b0, mag_fmap2b0, [("out_matrix_file", "in_matrix_file")]),
                (bias_mag_fmap, mag_fmap2b0, [("output_image", "in_file")]),
                (mask_ref_b0, mag_fmap2b0, [("out_file", "reference")]),
                # Apply the transformation on the calibrated fmap
                (bet_mag_fmap2b0, fmap2b0, [("out_matrix_file", "in_matrix_file")]),
                (calibrate_fmap, fmap2b0, [("output_node.calibrated_fmap", "in_file")]),
                (mask_ref_b0, fmap2b0, [("out_file", "reference")]),
                # # Smooth the registered (calibrated) fmap
                (fmap2b0, smoothing, [("out_file", "in_file")]),

                # Step 3: Run FSL eddy
                # ====================
                (init_node, eddy, [("dwi", "in_file"),
                                   ("bval", "in_bval"),
                                   ("bvec", "in_bvec"),
                                   ("image_id", "out_base")]),
                (gen_acq_txt, eddy, [("out_acq", "in_acqp")]),
                (gen_index_txt, eddy, [("out_index", "in_index")]),
                (smoothing, eddy, [("out_file", "field")]),
                (pre_mask_b0, eddy, [("out_file", "in_mask")]),

                # Step 4: Bias correction
                # =======================
                (init_node, bias, [("bval", "in_bval")]),
                (eddy, bias, [("out_rotated_bvecs", "in_bvec"),
                              ("out_corrected", "in_file")]),
                # Step 5: Final brainmask
                # =======================
                # Compute average b0 on corrected dataset (for brain mask extraction)
                (init_node, compute_avg_b0, [("bval", "in_bval")]),
                (bias, compute_avg_b0, [("out_file", "in_dwi")]),
                # Compute b0 mask on corrected avg b0
                (compute_avg_b0, mask_avg_b0, [("out_b0_average", "in_file")]),

                # Print end message
                (init_node, print_end_message, [("image_id", "image_id")]),
                (mask_avg_b0, print_end_message, [("mask_file", "final_file")]),

                # Output node
                (init_node, self.output_node, [("bval", "preproc_bval")]),
                (eddy, self.output_node, [("out_rotated_bvecs", "preproc_bvec")]),
                (bias, self.output_node, [("out_file", "preproc_dwi")]),
                (mask_avg_b0, self.output_node, [("mask_file", "b0_mask")]),
                (bet_mag_fmap2b0, self.output_node, [("out_file", "magnitude_on_b0")]),
                (fmap2b0, self.output_node, [("out_file", "calibrated_fmap_on_b0")]),
                (smoothing, self.output_node, [("out_file", "smoothed_fmap_on_b0")]),
            ]
        )
        # fmt: on



import typing as ty
from os import PathLike
from pathlib import PurePath

import pydra
from nipype.interfaces.ants import N4BiasFieldCorrection, RegistrationSynQuick
from pydra import Workflow
from pydra.mark import annotate, task

from clinica.pydra.engine import clinica_io

n4_bias_field_correction = N4BiasFieldCorrection(bspline_fitting_distance=300)
registration_syn_quick = RegistrationSynQuick(transform_type="a")


@task
@annotate({"return": {"cropped_image": PurePath}})
def crop_image(input_image: PathLike, template_image: PathLike) -> PurePath:
    from pathlib import Path

    from nilearn.image import resample_to_img

    cropped_image = Path(input_image).name.replace(".nii.gz", "_cropped.nii.gz")

    resample_to_img(
        source_img=str(input_image), target_img=str(template_image), force_resample=True
    ).to_filename(cropped_image)

    return cropped_image


def download_file(url: str, to: str) -> PurePath:
    from shutil import copyfileobj
    from ssl import SSLContext
    from urllib.request import urlopen

    print(f"Downloading {url} to {to}...")

    response = urlopen(url=url, context=SSLContext())
    with open(to, mode="wb") as f:
        copyfileobj(response, f)

    return PurePath(to)


@task
@annotate({"return": {"mni_template_file": PurePath}})
def download_mni_template() -> PurePath:
    from pathlib import Path

    return download_file(
        url="https://aramislab.paris.inria.fr/files/data/img_t1_linear/mni_icbm152_t1_tal_nlin_sym_09c.nii.gz",
        to=str(Path.cwd() / "mni_icbm152_t1_tal_nlin_sym_09c.nii.gz"),
    )


@task
@annotate({"return": {"ref_template_file": PurePath}})
def download_ref_template() -> PurePath:
    from pathlib import Path

    return download_file(
        url="https://aramislab.paris.inria.fr/files/data/img_t1_linear/ref_cropped_template.nii.gz",
        to=str(Path.cwd() / "ref_cropped_template.nii.gz"),
    )


@clinica_io
def build_core_workflow(name: str = "core", parameters={}) -> Workflow:
    """Core workflow for the T1 linear pipeline.

    Parameters
    ----------
    name : str
        The name of the workflow.

    Returns
    -------
    Workflow
        The core workflow.
    """
    from pydra.tasks.nipype1.utils import Nipype1Task

    input_spec = pydra.specs.SpecInfo(
        name="Input",
        fields=[("T1w", str, {"mandatory": True})],
        bases=(pydra.specs.BaseSpec,),
    )

    wf = Workflow(name, input_spec=input_spec)

    wf.add(download_mni_template(name="download_mni_template"))

    wf.add(download_ref_template(name="download_ref_template"))

    wf.add(
        Nipype1Task(
            name="n4_bias_field_correction",
            interface=n4_bias_field_correction,
            input_image=wf.lzin.T1w,
        ).split("input_image")
    )

    wf.add(
        Nipype1Task(
            name="registration_syn_quick",
            interface=registration_syn_quick,
            fixed_image=wf.download_mni_template.lzout.mni_template_file,
            moving_image=wf.n4_bias_field_correction.lzout.output_image,
        )
    )

    wf.add(
        crop_image(
            name="crop_image",
            interface=crop_image,
            input_image=wf.registration_syn_quick.lzout.warped_image,
            template_image=wf.download_ref_template.lzout.ref_template_file,
        )
    )

    wf.set_output(
        [
            ("t1w_corrected_file", wf.n4_bias_field_correction.lzout.output_image),
            ("t1w_registered_file", wf.registration_syn_quick.lzout.warped_image),
            ("t1w_cropped_file", wf.crop_image.lzout.cropped_image),
            ("xfm_file", wf.registration_syn_quick.lzout.out_matrix),
        ]
    )

    return wf



from typing import Optional

import click

from clinica.pipelines import cli_param
from clinica.pipelines.cli import cli as run_cli

pipeline_name = "t1-freesurfer-template"


@click.command(name=pipeline_name)
@cli_param.argument.caps_directory
@cli_param.option_group.common_pipelines_options
@cli_param.option.subjects_sessions_tsv
@cli_param.option.working_directory
@cli_param.option.n_procs
@cli_param.option.overwrite_outputs
def cli(
    caps_directory: str,
    subjects_sessions_tsv: Optional[str] = None,
    working_directory: Optional[str] = None,
    n_procs: Optional[int] = None,
    overwrite_outputs: bool = False,
) -> None:
    """Creation of unbiased template with FreeSurfer.

    https://aramislab.paris.inria.fr/clinica/docs/public/latest/Pipelines/T1_FreeSurfer_Longitudinal/
    """
    from networkx import Graph

    from clinica.utils.ux import print_end_pipeline

    from .t1_freesurfer_template_pipeline import T1FreeSurferTemplate

    pipeline = T1FreeSurferTemplate(
        caps_directory=caps_directory,
        tsv_file=subjects_sessions_tsv,
        base_dir=working_directory,
        name=pipeline_name,
        overwrite_caps=overwrite_outputs,
    )

    exec_pipeline = (
        pipeline.run(plugin="MultiProc", plugin_args={"n_procs": n_procs})
        if n_procs
        else pipeline.run()
    )

    if isinstance(exec_pipeline, Graph):
        print_end_pipeline(
            pipeline_name, pipeline.base_dir, pipeline.base_dir_was_specified
        )


run_cli.add_command(cli)

if __name__ == "__main__":
    cli()



from ..pydra.t1_linear import t1_linear_cli as pydra_t1_linear_cli  # noqa
from ..pydra.t1_volume.create_dartel import cli as pydra_t1vol_cd_cli  # noqa
from ..pydra.t1_volume.tissue_segmentation import cli as pydra_t1vol_ts_cli  # noqa
from . import (
    deeplearning_prepare_data,
    dwi_connectome,
    dwi_dti,
    dwi_preprocessing_using_fmap,
    dwi_preprocessing_using_t1,
    machine_learning,
    machine_learning_spatial_svm,
    pet_linear,
    pet_surface,
    pet_volume,
    statistics_surface,
    statistics_volume,
    statistics_volume_correction,
    t1_freesurfer,
    t1_freesurfer_longitudinal,
    t1_linear,
    t1_volume,
    t1_volume_create_dartel,
    t1_volume_dartel2mni,
    t1_volume_existing_template,
    t1_volume_parcellation,
    t1_volume_register_dartel,
    t1_volume_tissue_segmentation,
)



def init_input_node(caps_dir, participant_id, session_id, long_id, output_dir):
    """Initialize the pipeline."""
    import os
    import platform
    from tempfile import mkdtemp

    from clinica.utils.longitudinal import read_sessions
    from clinica.utils.stream import cprint
    from clinica.utils.ux import print_begin_image

    # Extract <image_id>
    image_id = "{0}_{1}_{2}".format(participant_id, session_id, long_id)

    # Create SUBJECTS_DIR for recon-all (otherwise, the command won't run)
    if platform.system().lower().startswith("darwin"):
        # Special case: On macOS, 'recon-all -long' can failed if the $SUBJECTS_DIR is too long
        # To circumvent this issue, we create a sym link in $(TMP) so that $SUBJECTS_DIR is a short path
        subjects_dir = mkdtemp()
        cprint(
            msg=(
                f"Needs to create a $SUBJECTS_DIR folder "
                f"in {subjects_dir} for {image_id.replace('_', ' | ')} (macOS case)."
            ),
            lvl="warning",
        )
    else:
        subjects_dir = os.path.join(output_dir, image_id)

    os.makedirs(subjects_dir, exist_ok=True)

    # Create symbolic link containing cross-sectional segmentation(s) in SUBJECTS_DIR so that recon-all can run
    for s_id in read_sessions(caps_dir, participant_id, long_id):
        cross_sectional_path = os.path.join(
            caps_dir,
            "subjects",
            participant_id,
            s_id,
            "t1",
            "freesurfer_cross_sectional",
            f"{participant_id}_{s_id}",
        )
        os.symlink(
            cross_sectional_path,
            os.path.join(subjects_dir, f"{participant_id}_{s_id}"),
        )

    # Create symbolic links containing unbiased template in SUBJECTS_DIR so that recon-all can run
    template_path = os.path.join(
        caps_dir,
        "subjects",
        participant_id,
        long_id,
        "freesurfer_unbiased_template",
        f"{participant_id}_{long_id}",
    )
    os.symlink(template_path, os.path.join(subjects_dir, f"{participant_id}_{long_id}"))

    print_begin_image(image_id)

    return subjects_dir


def run_recon_all_long(subjects_dir, participant_id, session_id, long_id, directive):
    """Run recon-all to create a longitudinal correction of a time point.

    Note:
    Longitudinal correction with FreeSurfer expects arguments to follow this syntax:
    recon-all -long <tpN_id> <template_id> -all; e.g.: recon-all -long sub-CLNC01_ses-M00 sub-CLNC01_long-M00M18 -all

    Currently, Nipype does not provide interface for "recon-all -long" case. As a result, ReconAll interface should be
    modified to handle this situation. In the meantime, the arguments of this function follows ReconAll.inputs names
    namely:
        - "-long <tpN_id> <template_id>" is likely to be fed to ReconAll.inputs.args
        - "-all" will be fed to ReconAll.inputs.directive

    Folder containing the longitudinal correction has the following convention:
    <tpN_id>.long.<template_id>; e.g.: sub-CLNC01_ses-M00.long.sub-CLNC01_long-M00M18
    which is automatically generated by FreeSurfer.

    This folder name is likely to be retrieved in ReconAll.outputs.subject_id.

    See official documentation (https://surfer.nmr.mgh.harvard.edu/fswiki/LongitudinalProcessing) for details.
    """
    import subprocess

    # Prepare arguments for recon-all.
    flags = " -long {0}_{1} {0}_{2} ".format(participant_id, session_id, long_id)

    recon_all_long_command = "recon-all {0} -sd {1} {2}".format(
        flags, subjects_dir, directive
    )
    subprocess_run_recon_all_long = subprocess.run(
        recon_all_long_command,
        shell=True,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )
    if subprocess_run_recon_all_long.returncode != 0:
        raise ValueError("recon-all -long failed, returned non-zero code")

    subject_id = "{0}_{1}.long.{0}_{2}".format(participant_id, session_id, long_id)

    return subject_id


def write_tsv_files(subjects_dir, subject_id):
    """Generate statistics TSV files in `subjects_dir`/regional_measures folder for `subject_id`.

    Notes:
        We do not need to check the line "finished without error" in scripts/recon-all.log.
        If an error occurs, it will be detected by Nipype and the next nodes (i.e.
        write_tsv_files will not be called).
    """
    import os

    from clinica.utils.freesurfer import (
        extract_image_id_from_longitudinal_segmentation,
        generate_regional_measures,
    )
    from clinica.utils.stream import cprint

    image_id = extract_image_id_from_longitudinal_segmentation(subject_id)
    str_image_id = (
        image_id.participant_id + "_" + image_id.session_id + "_" + image_id.long_id
    )
    if os.path.isfile(os.path.join(subjects_dir, subject_id, "mri", "aparc+aseg.mgz")):
        generate_regional_measures(subjects_dir, subject_id)
    else:
        cprint(
            msg=(
                f"{str_image_id.replace('_', ' | ')} does not contain mri/aseg+aparc.mgz file. "
                "Creation of regional_measures/ folder will be skipped."
            ),
            lvl="warning",
        )
    return subject_id


def move_subjects_dir_to_source_dir(subjects_dir, source_dir, subject_id):
    """Move content of `subjects_dir`/`subject_id` to `source_dir`.

    This function will move content of `subject_id` if recon-all has run
    in $(TMP). This happens when FreeSurfer is run on macOS. Content of
    $(TMP)/`subject_id` is copied to `source_dir` before the deletion of $(TMP).

    Args:
        subjects_dir: $(TMP), if segmentation was performed on macOS,
            <base_dir>/<Pipeline.Name>/ReconAll/`subject_id` otherwise
        source_dir: <base_dir>/<Pipeline.Name>/ReconAll folder
        subject_id: Subject ID (e.g. "sub-CLNC01_ses-M00.long.sub-CLNC01_long-M00M18")

    Returns:
        subject_id for node connection with Nipype
    """
    import os
    import shutil

    from clinica.utils.freesurfer import extract_image_id_from_longitudinal_segmentation
    from clinica.utils.stream import cprint

    image_id = extract_image_id_from_longitudinal_segmentation(subject_id)
    str_image_id = (
        image_id.participant_id + "_" + image_id.session_id + "_" + image_id.long_id
    )

    if source_dir not in subjects_dir:
        shutil.copytree(
            src=os.path.join(subjects_dir, subject_id),
            dst=os.path.join(source_dir, str_image_id, subject_id),
            symlinks=True,
        )
        shutil.copytree(
            src=os.path.join(subjects_dir, "regional_measures"),
            dst=os.path.join(source_dir, str_image_id, "regional_measures"),
            symlinks=True,
        )
        shutil.rmtree(subjects_dir)
        cprint(
            msg=(
                f"Segmentation of {subject_id.replace('_', ' | ')} "
                f"has moved to working directory and $SUBJECTS_DIR folder ({subjects_dir}) was deleted."
            ),
            lvl="warning",
        )

    return subject_id


def save_to_caps(source_dir, subject_id, caps_dir, overwrite_caps=False):
    """Save `source_dir`/`subject_id`/ to CAPS folder.

    This function copies FreeSurfer segmentation and regional_measures folder of `source_dir`/`image_id`/ to
    `caps_dir`/subjects/<participant_id>/<session_id>/<long_id>/t1_freesurfer_longitudinal/
    where `image_id` = <participant_id>_<session_id>_<long_id>.

    The `source_dir`/`image_id`/ folder should contain the following elements:
        - fsaverage, lh.EC_average and rh.EC_average symbolic links automatically generated by recon-all
        - symbolic links to cross-sectional segmentation(s) and unbiased template needed for recon-all
        - <participant_id>_<session_id>.long.<participant_id>_<long_id>/ folder containing the FreeSurfer segmentation
        - regional_measures/ folder containing TSV files

    Notes:
        We do not need to check the line "finished without error" in scripts/recon-all.log.
        If an error occurs, it will be detected by Nipype and the next nodes (i.e. save_to_caps will not be called).
    """
    import os
    import shutil

    from clinica.utils.freesurfer import extract_image_id_from_longitudinal_segmentation
    from clinica.utils.stream import cprint
    from clinica.utils.ux import print_end_image

    image_id = extract_image_id_from_longitudinal_segmentation(subject_id)
    participant_id = image_id.participant_id
    session_id = image_id.session_id
    long_id = image_id.long_id
    str_image_id = (
        image_id.participant_id + "_" + image_id.session_id + "_" + image_id.long_id
    )

    destination_dir = os.path.join(
        os.path.expanduser(caps_dir),
        "subjects",
        participant_id,
        session_id,
        "t1",
        long_id,
        "freesurfer_longitudinal",
    )

    # Save FreeSurfer segmentation
    representative_file = os.path.join(subject_id, "mri", "aparc+aseg.mgz")
    representative_source_file = os.path.join(
        os.path.expanduser(source_dir), str_image_id, representative_file
    )
    representative_destination_file = os.path.join(destination_dir, representative_file)
    if os.path.isfile(representative_source_file):
        if os.path.isfile(representative_destination_file):
            if overwrite_caps:
                shutil.rmtree(destination_dir)
                shutil.copytree(
                    src=os.path.join(source_dir, subject_id, subject_id),
                    dst=os.path.join(destination_dir, subject_id),
                    symlinks=True,
                )
                shutil.copytree(
                    src=os.path.join(source_dir, subject_id, "regional_measures"),
                    dst=os.path.join(destination_dir, "regional_measures"),
                    symlinks=True,
                )
        else:
            shutil.copytree(
                src=os.path.join(source_dir, str_image_id, subject_id),
                dst=os.path.join(destination_dir, subject_id),
                symlinks=True,
            )
            shutil.copytree(
                src=os.path.join(source_dir, str_image_id, "regional_measures"),
                dst=os.path.join(destination_dir, "regional_measures"),
                symlinks=True,
            )
        print_end_image(str_image_id)
    else:
        cprint(
            msg=f"{subject_id.replace('_', ' | ')}  does not contain mri/aseg+aparc.mgz file. Copy will be skipped.",
            lvl="warning",
        )

    return str_image_id


def get_processed_images(caps_directory, part_ids, sess_ids, long_ids, atlas=False):
    """
    Extract image IDs (e.g. ["sub-CLNC01_ses-M00_long-M00M18", "sub-CLNC01_ses-M18_long-M00M18"]) of outputs
    already processed by T1FreeSurferLongitudinalCorrection pipeline.
    """
    import os

    image_ids = []
    if os.path.isdir(caps_directory):
        for (participant_id, session_id, long_id) in zip(part_ids, sess_ids, long_ids):
            output_file_pre = os.path.join(
                os.path.expanduser(caps_directory),
                "subjects",
                participant_id,
                session_id,
                "t1",
                long_id,
                "freesurfer_longitudinal",
                f"{participant_id}_{session_id}.long.{participant_id}_{long_id}",
            )
            if atlas:
                output_file = os.path.join(
                    output_file_pre, "stat", "rh.", atlas, ".stats"
                )
            else:
                output_file = os.path.join(
                    output_file_pre,
                    "mri",
                    "aparc+aseg.mgz",
                )
            if os.path.isfile(output_file):
                image_ids.append(f"{participant_id}_{session_id}_{long_id}")
    return image_ids



def prepare_phasediff_fmap(name="prepare_phasediff_fmap"):
    """This workflow adapts the fsl_prepare_fieldmap script from FSL for the FSL eddy command.

    Please note that the step 3 converts the fieldmap into Hz instead of rad/s
    in the initial script because FSL eddy --field expects the fieldmap to be in Hz.

    Input node:
        fmap_mask (str): Binary mask of the fieldmap.
        fmap_phasediff (str): Phase difference fieldmap.
        fmap_magnitude (str): Brain extracted magnitude fieldmap. Chose the fieldmap with the best contrast.
        delta_echo_time (float): DeltaEchoTime from BIDS specifications.

    Output node:
        calibrated_fmap (str): Calibrated fieldmap for eddy --field command.

    Warnings:
        This workflow can not be used for PRELUDE. You would need to change the step 3 (conversion into rad/s instead of Hz).
    """
    import nipype.interfaces.fsl as fsl
    import nipype.interfaces.utility as nutil
    import nipype.pipeline.engine as npe
    from niflow.nipype1.workflows.dmri.fsl.utils import (
        cleanup_edge_pipeline,
        demean_image,
        siemens2rads,
    )

    from .dwi_preprocessing_using_phasediff_fmap_utils import rads2hz

    input_node = npe.Node(
        nutil.IdentityInterface(
            fields=["fmap_mask", "fmap_phasediff", "fmap_magnitude", "delta_echo_time"]
        ),
        name="input_node",
    )

    output_node = npe.Node(
        nutil.IdentityInterface(fields=["calibrated_fmap"]), name="output_node"
    )

    # Step 1 - Convert the fmap into radians
    pha2rads = npe.Node(
        nutil.Function(
            input_names=["in_file"], output_names=["out_file"], function=siemens2rads
        ),
        name="1-ConvertFMapToRads",
    )

    # Step 2 - Unwrap the fmap with PRELUDE
    prelude = npe.Node(fsl.PRELUDE(process3d=True), name="2-PhaseUnwrap")

    # Step 3 - Convert the fmap to Hz
    rads2hz = npe.Node(
        nutil.Function(
            input_names=["in_file", "delta_te"],
            output_names=["out_file"],
            function=rads2hz,
        ),
        name="3-ConvertFMapToHz",
    )

    # Step 4 - Call FUGUE to extrapolate from mask (fill holes, etc)
    pre_fugue = npe.Node(fsl.FUGUE(save_fmap=True), name="4-FugueExtrapolationFromMask")

    # Step 5 - Demean fmap to avoid gross shifting
    demean = npe.Node(
        nutil.Function(
            input_names=["in_file", "in_mask"],
            output_names=["out_file"],
            function=demean_image,
        ),
        name="5-DemeanFMap",
    )

    # Step 6 - Clean up edge voxels
    cleanup = cleanup_edge_pipeline(name="6-CleanUpEdgeVoxels")

    wf = npe.Workflow(name=name)
    # fmt: off
    wf.connect(
        [
            # Step 1 - Convert the fmap into radians
            (input_node, pha2rads, [("fmap_phasediff", "in_file")]),
            # Step 2 - Unwrap the fmap with PRELUDE
            (pha2rads, prelude, [("out_file", "phase_file")]),
            (input_node, prelude, [("fmap_magnitude", "magnitude_file")]),
            (input_node, prelude, [("fmap_mask", "mask_file")]),
            # Step 3 - Convert the fmap to Hz
            (prelude, rads2hz, [("unwrapped_phase_file", "in_file")]),
            (input_node, rads2hz, [("delta_echo_time", "delta_te")]),
            # Step 4 - Call FUGUE to extrapolate from mask (fill holes, etc)
            (rads2hz, pre_fugue, [("out_file", "fmap_in_file")]),
            (input_node, pre_fugue, [("fmap_mask", "mask_file")]),
            # Step 5 - Demean fmap to avoid gross shifting
            (pre_fugue, demean, [("fmap_out_file", "in_file")]),
            (input_node, demean, [("fmap_mask", "in_mask")]),
            # Step 6 - Clean up edge voxels
            (demean, cleanup, [("out_file", "inputnode.in_file")]),
            (input_node, cleanup, [("fmap_mask", "inputnode.in_mask")]),
            # Output node
            (cleanup, output_node, [("outputnode.out_file", "calibrated_fmap")]),
        ]
    )
    # fmt: on

    return wf



import clinica.pipelines.engine as cpe


class StatisticsSurface(cpe.Pipeline):
    """StatisticsSurface - Surface-based mass-univariate analysis with SurfStat.

    See documentation at https://aramislab.paris.inria.fr/clinica/docs/public/latest/Pipelines/Stats_Surface/

    Note:
        The `tsv_file` attribute is overloaded for this pipeline. It must contain a list of subjects
        with their sessions and all the covariates and factors needed for the GLM.

        Pipeline parameters are explained in StatisticsSurfaceCLI.define_options()

    Returns:
        A clinica pipeline object containing the StatisticsSurface pipeline.
    """

    def check_pipeline_parameters(self):
        """Check pipeline parameters."""
        from clinica.utils.exceptions import ClinicaException
        from clinica.utils.group import check_group_label

        from ._inputs import _get_t1_freesurfer_custom_file_template

        # Clinica compulsory parameters
        self.parameters.setdefault("group_label", None)
        check_group_label(self.parameters["group_label"])

        if "orig_input_data" not in self.parameters.keys():
            raise KeyError(
                "Missing compulsory orig_input_data key in pipeline parameter."
            )

        self.parameters.setdefault("glm_type", None)
        if self.parameters["glm_type"] not in ["group_comparison", "correlation"]:
            raise ClinicaException(
                f"The glm_type you specified is wrong: it should be group_comparison or "
                f"correlation (given value: {self.parameters['glm_type']})."
            )

        if "contrast" not in self.parameters.keys():
            raise KeyError("Missing compulsory contrast key in pipeline parameter.")

        # Optional parameters
        self.parameters.setdefault("covariates", None)
        self.parameters.setdefault("full_width_at_half_maximum", 20)

        # Optional parameters for inputs from pet-surface pipeline
        self.parameters.setdefault("acq_label", None)
        self.parameters.setdefault("suvr_reference_region", None)

        # Optional parameters for custom pipeline
        self.parameters.setdefault(
            "custom_file", _get_t1_freesurfer_custom_file_template(self.base_dir)
        )
        self.parameters.setdefault("measure_label", "ct")

        # Advanced parameters
        self.parameters.setdefault("cluster_threshold", 0.001)
        if (
            self.parameters["cluster_threshold"] < 0
            or self.parameters["cluster_threshold"] > 1
        ):
            raise ClinicaException(
                f"Cluster threshold should be between 0 and 1 "
                f"(given value: {self.parameters['cluster_threshold']})."
            )

    def check_custom_dependencies(self):
        """Check dependencies that can not be listed in the `info.json` file."""

    def get_input_fields(self):
        """Specify the list of possible inputs of this pipeline.

        Returns:
            A list of (string) input fields name.
        """
        return []

    def get_output_fields(self):
        """Specify the list of possible outputs of this pipeline.

        Returns:
            A list of (string) output fields name.
        """
        return ["output_dir"]

    def build_input_node(self):
        """Build and connect an input node to the pipeline."""
        import os

        from clinica.utils.exceptions import ClinicaException
        from clinica.utils.inputs import clinica_file_reader
        from clinica.utils.stream import cprint

        # Check if already present in CAPS
        # ================================
        # Check if the group label has been existed, if yes, give an error to the users
        # Note(AR): if the user wants to compare Cortical Thickness measure with PET measure
        # using the group_id, Clinica won't allow it.
        # TODO: Modify this behaviour
        if os.path.exists(
            os.path.join(
                self.caps_directory, "groups", f"group-{self.parameters['group_label']}"
            )
        ):
            raise ClinicaException(
                f"Group label {self.parameters['group_label']} already exists, "
                f"please choose another one or delete the existing folder and "
                f"also the working directory and rerun the pipeline"
            )

        # Check input files before calling SurfStat with Matlab
        # =====================================================
        all_errors = []
        # clinica_files_reader expects regexp to start at subjects/ so sub-*/ses-*/ is removed here
        fwhm = str(self.parameters["full_width_at_half_maximum"])
        for direction, hemi in zip(["left", "right"], ["lh", "rh"]):
            cut_pattern = "sub-*/ses-*/"
            query = {"subject": "sub-*", "session": "ses-*", "hemi": hemi, "fwhm": fwhm}
            pattern_hemisphere = self.parameters["custom_file"] % query
            surface_based_info = {
                "pattern": pattern_hemisphere[
                    pattern_hemisphere.find(cut_pattern) + len(cut_pattern) :
                ],
                "description": f"surface-based features on {direction} hemisphere at FWHM = {fwhm}",
            }
            try:
                clinica_file_reader(
                    self.subjects,
                    self.sessions,
                    self.caps_directory,
                    surface_based_info,
                )
            except ClinicaException as e:
                all_errors.append(e)
        # Raise all errors if something happened
        if len(all_errors) > 0:
            error_message = "Clinica faced errors while trying to read files in your CAPS directory.\n"
            for msg in all_errors:
                error_message += str(msg)
            raise RuntimeError(error_message)

        # Give pipeline info
        # ==================
        cprint(
            "The pipeline will last a few minutes. Images generated by Matlab will popup during the pipeline."
        )

    def build_output_node(self):
        """Build and connect an output node to the pipeline."""
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from .statistics_surface_utils import save_to_caps

        # Writing results into CAPS
        # =========================
        save_to_caps = npe.Node(
            interface=nutil.Function(
                input_names=[
                    "source_dir",
                    "caps_dir",
                    "overwrite_caps",
                    "pipeline_parameters",
                ],
                function=save_to_caps,
            ),
            name="SaveToCaps",
        )
        save_to_caps.inputs.caps_dir = self.caps_directory
        save_to_caps.inputs.overwrite_caps = self.overwrite_caps
        save_to_caps.inputs.pipeline_parameters = self.parameters

        self.connect(
            [
                (self.output_node, save_to_caps, [("output_dir", "source_dir")]),
            ]
        )

    def build_core_nodes(self):
        """Build and connect the core nodes of the pipeline."""
        import os

        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        import clinica.pipelines.statistics_surface.statistics_surface_utils as utils

        from .statistics_surface_utils import init_input_node

        init_input = npe.Node(
            interface=nutil.Function(
                input_names=["parameters", "base_dir", "subjects_visits_tsv"],
                output_names=["group_label", "surfstat_results_dir"],
                function=init_input_node,
            ),
            name="0-InitPipeline",
        )
        init_input.inputs.parameters = self.parameters
        init_input.inputs.base_dir = os.path.join(self.base_dir, self.name)
        init_input.inputs.subjects_visits_tsv = self.tsv_file

        # Node to wrap the SurfStat matlab script
        surfstat = npe.Node(
            name="1-RunSurfStat",
            interface=nutil.Function(
                input_names=[
                    "caps_dir",
                    "output_dir",
                    "subjects_visits_tsv",
                    "pipeline_parameters",
                ],
                output_names=["output_dir"],
                function=utils.run_matlab,
            ),
        )
        surfstat.inputs.caps_dir = self.caps_directory
        surfstat.inputs.subjects_visits_tsv = self.tsv_file
        surfstat.inputs.pipeline_parameters = self.parameters

        # Connection
        # ==========
        self.connect(
            [
                (init_input, surfstat, [("surfstat_results_dir", "output_dir")]),
                (surfstat, self.output_node, [("output_dir", "output_dir")]),
            ]
        )



import xml.etree.ElementTree
from pathlib import Path
from typing import Optional, Tuple, Union

import pandas as pd

# Map from names of extracted metadata to their proper BIDS names.
METADATA_NAME_MAPPING = {
    "acquisition_type": "MRAcquisitionType",
    "pulse_sequence": "PulseSequenceType",
    "manufacturer": "Manufacturer",
    "field_strength": "MagneticFieldStrength",
}


def _read_xml_files(subj_ids: Optional[list] = None, xml_path: str = "") -> list:
    """Return the XML files in the folder `xml_path` for the provided `subj_ids`.
    This function assumes that file are named "ADNI_{sub_ids}.xml".
    If no files were found, an `IndexError` is raised.
    """
    from glob import glob
    from os import path

    if subj_ids:
        xml_files = []
        xml_regex = [path.join(xml_path, ("ADNI_" + e + "*.xml")) for e in subj_ids]
        for subj_files in xml_regex:
            xml_files.extend(glob(subj_files))
    else:
        xml_files = glob("Clinica_processed_metadata/ADNI_*.xml")
    if len(xml_files) == 0:
        raise IndexError("No ADNI xml files were found for reading the metadata.")
    return xml_files


def _check_xml_tag(el_tag: str, exp_tag: str):
    """Check that the XML element tage matches the expected tag."""
    if el_tag != exp_tag:
        raise ValueError(f"Bad tag: expected {exp_tag}, got {el_tag}")


def _check_xml_nb_children(xml_el: xml.etree.ElementTree.Element, exp_nb_children: int):
    """Check that the given XML element has the expected number of children.
    Raise a `ValueError` otherwise.
    """
    nb_children = len(xml_el)
    if isinstance(exp_nb_children, int):
        if nb_children != exp_nb_children:
            raise ValueError(
                f"Bad number of children for <{xml_el.tag}>: "
                f"got {nb_children} != {exp_nb_children}"
            )
    else:
        # container
        if nb_children not in exp_nb_children:
            raise ValueError(
                f"Bad number of children for <{xml_el.tag}>: "
                f"got {nb_children}, not in {exp_nb_children}"
            )


def _check_xml(
    xml_el: xml.etree.ElementTree.Element, exp_tag: str, exp_nb_children=None
) -> xml.etree.ElementTree.Element:
    """Check that the given xml element is as expected (tag name + nb of children).
    Raise a ValueError if not and return the element if it is valid.
    """
    _check_xml_tag(xml_el.tag, exp_tag)
    if exp_nb_children is not None:  # no check otherwise
        _check_xml_nb_children(xml_el, exp_nb_children)
    return xml_el


def _parse_project(
    root: xml.etree.ElementTree.Element,
) -> xml.etree.ElementTree.Element:
    """Check that root has 1 child: <project> having 4 children,
    extract project from them.
    Check that the project identifier contains ADNI.
    """
    if len(root) != 1:
        raise ValueError(
            "XML root should have only one child. " f"{len(root)} children found."
        )
    project = _check_xml(root[0], "project", 4)
    _check_xml_project_identifier(project, "ADNI")
    return project


def _check_xml_and_get_text(
    xml_el: xml.etree.ElementTree.Element, exp_tag: str, *, cast=None
) -> str:
    """Check xml element and return its text (leaf)"""
    _check_xml(xml_el, exp_tag, 0)  # leaf
    return _get_text(xml_el, cast=cast)


def _get_text(xml_el: xml.etree.ElementTree.Element, cast=None) -> str:
    """Get text of xml element (leaf)."""
    if cast is not None:
        return cast(xml_el.text)
    return xml_el.text


def _check_derived_image_xml(derived: xml.etree.ElementTree.Element):
    """Perform sanity checks on derived images."""
    assert len(derived) >= 9  # multiple <provenanceDetail> nodes possible
    assert (
        _check_xml_and_get_text(derived[2], "imageType") == "image volume"
    ), "imageType"
    assert _check_xml_and_get_text(derived[3], "tissue") == "All", "tissue"
    assert _check_xml_and_get_text(derived[4], "hemisphere") == "Both", "hemis"
    assert (
        _check_xml_and_get_text(derived[5], "anatomicStructure") == "Brain"
    ), "anat struct"
    assert (
        _check_xml_and_get_text(derived[6], "registration") == "native"
    ), "registration not native"
    # skip check relatedImage, original ID consistence + "derived from"...
    # skip processing steps with names, versions, dates and so on
    # 2 cases
    if derived[-1].tag == "creationDate":
        assert (
            _check_xml_and_get_text(derived[-1], "creationDate") == "0000-00-00"
        ), "creaDate"
    else:
        assert (
            _check_xml_and_get_text(derived[-2], "creationDate") == "0000-00-00"
        ), "creaDate"


def _get_derived_image_metadata(derived: xml.etree.ElementTree.Element) -> dict:
    """Return metadata for derived images after
    performing some sanity checks.
    """
    _check_derived_image_xml(derived)
    return {
        "image_proc_id": _check_xml_and_get_text(derived[0], "imageUID", cast=int),
        "image_proc_desc": _check_xml_and_get_text(
            derived[1], "processedDataLabel"
        ).strip(),
    }


def _check_xml_project_identifier(
    project: xml.etree.ElementTree.Element, expected: str
):
    """Check the project identifier."""
    if _check_xml_and_get_text(project[0], "projectIdentifier") != expected:
        raise ValueError(f"Not {expected} cohort")


def _get_original_image_metadata(original_image: xml.etree.ElementTree.Element) -> dict:
    """Get original image metadata."""
    if not isinstance(original_image, list):
        return _get_image_metadata(original_image)
    # only keep common metadata (remaining will be None) and
    original_images_metadata = [_get_image_metadata(_) for _ in original_image]
    original_images_metadata = {
        _.pop("image_orig_id"): _ for _ in original_images_metadata
    }
    original_image_metadata = {
        "image_orig_id": "|".join(map(str, original_images_metadata.keys()))
    }
    for k in next(iter(original_images_metadata.values())).keys():
        # loop on original images metadata and only add consistent metadata
        set_v = set(_[k] for _ in original_images_metadata.values())
        if len(set_v) == 1:
            original_image_metadata[k] = set_v.pop()
    return original_image_metadata


def _check_processed_image_rating(
    processed_image_rating: Union[int, None],
    original_image_metadata: dict,
    derived_image_metadata: dict,
):
    """Check, if possible, that the rating of the processed images are
    consistent with the rating of the original image.
    """
    from clinica.utils.stream import cprint

    rating = original_image_metadata.get("image_orig_rating", None)
    if processed_image_rating is not None:
        if processed_image_rating != rating:
            cprint(
                msg=(
                    "Image rating for processed image "
                    f"{derived_image_metadata.get('image_proc_id')} not "
                    "consistent with rating of original image"
                ),
                lvl="info",
            )


def _get_root_from_xml_path(xml_path: str) -> xml.etree.ElementTree.Element:
    """Return the root XML element from the XML file path."""
    import os
    from xml.etree import ElementTree

    try:
        tree = ElementTree.parse(xml_path)
        root = tree.getroot()
    except Exception as e:
        raise ValueError(os.path.basename(xml_path) + ": " + str(e))
    return root


def _parse_series(
    study: xml.etree.ElementTree.Element,
) -> xml.etree.ElementTree.Element:
    """Get series xml element from study xml element."""
    expected_length = 3 if len(study) == 7 else 4
    return _check_xml(study[5], "series", expected_length)


def _parse_derived_images(
    series: xml.etree.ElementTree.Element,
) -> Tuple[Union[xml.etree.ElementTree.Element, list], Union[int, None], dict]:
    """Parse the series xml element when derived images are present.
    Return the original image, the rating of the processed images,
    and the derived image metadata as a dict.
    """
    # we can have 1 or 2 related imgs...
    series_meta = _check_xml(series[-1], "seriesLevelMeta", {3, 4})
    related_image = _check_xml(series_meta[2], "relatedImageDetail", 1)
    original_image = _check_xml(related_image[0], "originalRelatedImage", {3, 4})
    if len(series_meta) == 4:
        # 2 original images
        related_image_bis = _check_xml(series_meta[-1], "relatedImageDetail", 1)
        original_image_bis = _check_xml(
            related_image_bis[0], "originalRelatedImage", len(original_image)
        )
        original_image = [original_image, original_image_bis]
    # specific
    pipeline_name = _check_xml_and_get_text(
        _check_xml(series_meta[0], "annotation", 1)[0], "text"
    )
    deriv = _check_xml(series_meta[1], "derivedProduct")
    derived_metadata = {"image_proc_pipe": pipeline_name.strip()}
    derived_metadata.update(_get_derived_image_metadata(deriv))
    processed_image_rating = _get_image_rating(deriv[-1])
    return original_image, processed_image_rating, derived_metadata


def _check_image(img: xml.etree.ElementTree.Element):
    """Check that the image XML element is valid."""
    if img.tag not in {
        "imagingProtocol",  # for original image metadata
        "originalRelatedImage",  # for processed image
    }:
        raise ValueError(
            f"Bad image tag <{img.tag}>. "
            "Should be either 'imagingProtocol' "
            "or 'originalRelatedImage'."
        )
    if len(img) not in {3, 4}:
        raise ValueError(
            f"Image XML element has {len(img)} children. " "Expected either 3 or 4."
        )


def _clean_protocol_metadata(protocol_metadata: dict) -> dict:
    """Replace confusing '|' (reserved for separation btw elements) with space.
    Only 1 manufacturer with this apparently but...
    """
    return {
        k: v.replace("|", " ")
        if isinstance(v, str)
        else v  # there are some None (not str)
        for k, v in protocol_metadata.items()
    }


def _filter_metadata(metadata: dict) -> dict:
    """Filter and clean a given metadata dictionary according to the
    METADATA_NAME_MAPPING dictionary."""
    filtered = dict()
    for k, v in metadata.items():
        if k in METADATA_NAME_MAPPING:
            filtered[METADATA_NAME_MAPPING[k]] = v
    return filtered


def _get_image_metadata(img: xml.etree.ElementTree.Element) -> dict:
    """Return information on original image as dict of metadata."""
    _check_image(img)
    protocol = _check_xml(img[2], "protocolTerm")  # do not check children
    assert all(p.tag == "protocol" and "term" in p.attrib.keys() for p in protocol)
    protocol_metadata = {
        "_".join(p.attrib["term"].split()).lower(): p.text for p in protocol
    }
    protocol_metadata = _clean_protocol_metadata(protocol_metadata)
    filtered_protocol_metadata = _filter_metadata(protocol_metadata)
    return {
        "image_orig_id": _check_xml_and_get_text(img[0], "imageUID", cast=int),
        "image_orig_seq": _check_xml_and_get_text(
            img[1], "description"
        ),  # .replace('|', ' '),
        **filtered_protocol_metadata,
    }


def _get_image_rating(image_rating: xml.etree.ElementTree.Element) -> Optional[str]:
    """Get the image rating value as an integer from the xml element."""
    if image_rating.tag != "imageRating":
        return None
    _check_xml(image_rating, "imageRating", 2)
    image_rating_desc = _check_xml_and_get_text(image_rating[0], "ratingDescription")
    image_rating_val = _check_xml_and_get_text(image_rating[1], "value", cast=int)
    assert image_rating_desc == str(image_rating_val)
    return image_rating_val


def _check_modality(study: xml.etree.ElementTree.Element, expected_modality: str):
    """Check that the modality of the given study is the expected one."""
    series = _parse_series(study)
    modality = _check_xml_and_get_text(series[1], "modality")
    if modality != expected_modality:
        raise ValueError(
            f"Unexpected modality {modality}, expected {expected_modality}."
        )


def _parse_subject(
    project: xml.etree.ElementTree.Element,
) -> Tuple[str, xml.etree.ElementTree.Element]:
    """From the project xml element, parse the subject and subject id."""
    subject = _check_xml(project[-1], "subject", {5, 7})  # with APOE or not
    subject_id = _check_xml_and_get_text(subject[0], "subjectIdentifier")
    return subject_id, subject


def _parse_study(
    subject: xml.etree.ElementTree.Element,
) -> xml.etree.ElementTree.Element:
    """Parse the study from the subject xml element
    and check that it is MRI.
    """
    study = _check_xml(subject[-1], "study", {6, 7})
    _check_modality(study, "MRI")
    return study


def _parse_images(
    study: xml.etree.ElementTree.Element,
) -> Tuple[Union[xml.etree.ElementTree.Element, list], Union[int, None], dict]:
    """Parse the images, original and derived if available."""
    if len(study) == 7:
        # Original only
        original_image = _check_xml(study[6], "imagingProtocol", {3, 4})
        return original_image, None, dict()
    else:
        series = _parse_series(study)
        return _parse_derived_images(series)


def _parse_xml_file(xml_path: str) -> dict:
    """Parse the given XML file and return the desired metadata as a dict."""
    root = _get_root_from_xml_path(xml_path)
    project = _parse_project(root)
    subject_id, subject = _parse_subject(project)
    study = _parse_study(subject)
    series = _parse_series(study)
    original_image, processed_image_rating, derived_image_metadata = _parse_images(
        study
    )
    original_image_metadata = _get_original_image_metadata(original_image)
    _check_processed_image_rating(
        processed_image_rating, original_image_metadata, derived_image_metadata
    )
    acquisition_date = pd.Timestamp(_check_xml_and_get_text(series[2], "dateAcquired"))
    scan_metadata = {
        "id": subject_id,
        "acq_time": acquisition_date,
        **original_image_metadata,
        **derived_image_metadata,
    }
    if "image_proc_id" not in scan_metadata and "image_orig_id" not in scan_metadata:
        raise ValueError(f"Scan metadata for subject {subject_id} has no image ID.")
    return scan_metadata


class FuncWithException:
    def __init__(self, f):
        self.f = f

    def __call__(self, *args, **kwargs):
        """Return a tuple: (result, exception)"""
        try:
            return self.f(*args, **kwargs), None
        except Exception as e:
            return None, e


def _run_parsers(xml_files: list) -> Tuple[list, dict]:
    """Run the parser `_parse_xml_file` on the list of files `xml_files`.
    Returns a tuple consisting if parsed images
    metadata and captured exceptions.
    """
    import os

    parser = FuncWithException(_parse_xml_file)
    imgs_with_excep = dict(
        zip(
            xml_files,  # tqdm is buggy when chunksize > 1
            [parser(xml_file) for xml_file in xml_files],
        )
    )
    imgs = [img for img, e in imgs_with_excep.values() if e is None]
    exceps = {
        os.path.basename(xml_p): e
        for xml_p, (_, e) in imgs_with_excep.items()
        if e is not None
    }
    return imgs, exceps


def _create_mri_meta_df(imgs: list) -> pd.DataFrame:
    """Create dataframe from images metadata parsed from the XML files."""
    mri_meta = pd.DataFrame(imgs).convert_dtypes()
    # Scan ID as expected by Clinica (derived if so, original otherwise)
    mri_meta["T1w_scan_id"] = mri_meta["image_proc_id"].combine_first(
        mri_meta["image_orig_id"]
    )
    # Index
    mri_meta.set_index(
        [
            "T1w_scan_id",
            "id",
            # "GENDER",
            # "VISIT_DESC",
            # "VISIT_AGE",
            # "SCAN_DATE",
            # "STUDY_ID",
            # "SERIES_ID",
        ],
        inplace=True,
    )
    mri_meta.reset_index(inplace=True)
    return mri_meta


def _get_existing_scan_dataframe(
    subj_path: str, session: str
) -> Union[Tuple[pd.DataFrame, Path], Tuple[None, Path]]:
    """Retrieve existing scan dataframe at the given
    `subj_path`, and the given `session`.
    If no existing scan file is found, a warning is given.
    """
    import warnings

    subj_id = Path(subj_path).name
    scans_tsv_path = Path(subj_path) / session / f"{subj_id}_{session}_scans.tsv"
    if scans_tsv_path.exists():
        df_scans = pd.read_csv(scans_tsv_path, sep="\t")
        df_scans["scan_id"] = df_scans["scan_id"].astype("Int64")
        return df_scans, scans_tsv_path
    warnings.warn(f"No scan tsv file for subject {subj_id} and session {session}")
    return None, scans_tsv_path


def _merge_scan_and_metadata(
    df_scans: pd.DataFrame, df_meta: pd.DataFrame, strategy: dict
) -> pd.DataFrame:
    """Perform a merge between the two provided dataframe
    according to the strategy.
    """
    return pd.merge(df_scans, df_meta, **strategy)


def _get_json_filename_from_scan_filename(scan_filename: Path) -> Path:
    """Replace the given `scan_filename` extensions with '.json'."""
    extensions = "".join(scan_filename.suffixes)
    return Path(str(scan_filename).replace(extensions, ".json"))


def _add_json_scan_metadata(
    json_path: Path,
    metadata: Union[dict, str],
    indent: int = 4,
    keep_none: bool = False,
):
    """Load existing metadata at the provided `json_path`, merge with new
    provided `metadata`, and finally write to disk at `json_path`.

    .. warning::
        This overwrites the existing json file.
    """
    import json

    existing_metadata = dict()
    if json_path.exists():
        with open(json_path, "r") as fp:
            existing_metadata = json.load(fp)
    if isinstance(metadata, str):
        metadata = json.loads(metadata)
    filtered_metadata = {
        k: v for k, v in metadata.items() if k in METADATA_NAME_MAPPING.values()
    }
    updated_metadata = {**existing_metadata, **filtered_metadata}
    if not keep_none:
        updated_metadata = {k: v for k, v in updated_metadata.items() if v is not None}
    if len(updated_metadata) > 0:
        with open(json_path, "w") as fp:
            json.dump(updated_metadata, fp, indent=indent)


def _add_metadata_to_scans(df_meta: pd.DataFrame, bids_subjs_paths: list) -> None:
    """Add the metadata to the appropriate tsv and json files."""
    from clinica.iotools.bids_utils import get_bids_sess_list

    merge_strategy = {"how": "left", "left_on": "scan_id", "right_on": "T1w_scan_id"}

    for subj_path in bids_subjs_paths:
        sess_list = get_bids_sess_list(subj_path)
        if sess_list:
            for sess in sess_list:
                df_scans, scans_tsv_path = _get_existing_scan_dataframe(subj_path, sess)
                if df_scans is not None:
                    columns_to_keep = list(df_scans.columns) + ["acq_time"]
                    df_merged = _merge_scan_and_metadata(
                        df_scans, df_meta, merge_strategy
                    )
                    for _, scan_row in df_merged.iterrows():
                        scan_path = Path(subj_path) / sess / scan_row["filename"]
                        json_path = _get_json_filename_from_scan_filename(scan_path)
                        _add_json_scan_metadata(json_path, scan_row.to_json())
                    df_merged[columns_to_keep].to_csv(scans_tsv_path, sep="\t")
    return None


def create_json_metadata(bids_subjs_paths: list, bids_ids: list, xml_path: str) -> None:
    """Create json metadata dictionary and add the metadata to the
    appropriate files in the BIDS hierarchy."""
    from clinica.iotools.converters.adni_to_bids.adni_utils import bids_id_to_loni

    loni_ids = [bids_id_to_loni(bids_id) for bids_id in bids_ids]
    xml_files = _read_xml_files(loni_ids, xml_path)
    imgs, exe = _run_parsers(xml_files)
    df_meta = _create_mri_meta_df(imgs)
    _add_metadata_to_scans(df_meta, bids_subjs_paths)
    return None



"""Common CLI arguments and options used by converters."""

import click

dataset_directory = click.argument(
    "dataset_directory",
    type=click.Path(exists=True, file_okay=False, resolve_path=True),
)

clinical_data_directory = click.argument(
    "clinical_data_directory",
    type=click.Path(exists=True, file_okay=False, resolve_path=True),
)

bids_directory = click.argument(
    "bids_directory",
    type=click.Path(writable=True, file_okay=False, resolve_path=True),
)

clinical_data_only = click.option(
    "-c",
    "--clinical-data-only",
    is_flag=True,
    help="Convert clinical data only.",
)

subjects_list = click.option(
    "-sl",
    "--subjects_list",
    type=click.Path(exists=True),
    help=(
        "Path to a text file containing a list of specific subjects to extract."
        "The expected format is one subject ID per line."
    ),
)



from enum import Enum
from typing import IO

from attrs import define, fields
from cattr.gen import make_dict_unstructure_fn, override
from cattr.preconf.json import make_converter

BIDS_VERSION = "1.7.0"


@define
class BIDSDatasetDescription:
    """Model representing a BIDS dataset description.

    See https://bids-specification.readthedocs.io/en/stable/03-modality-agnostic-files.html
    """

    class DatasetType(str, Enum):
        raw = "raw"
        derivative = "derivative"

    name: str
    bids_version: str = BIDS_VERSION
    dataset_type: DatasetType = DatasetType.raw

    def write(self, to: IO[str]):
        import json

        json.dump(converter.unstructure(self), to, indent=4)


def _rename(name: str) -> str:
    """Rename attributes following the specification for the JSON file.

    Basically pascal case with known acronyms such as BIDS fully capitalized.
    """
    return "".join(
        word.upper() if word == "bids" else word.capitalize()
        for word in name.split("_")
    )


# Register a JSON converter for the BIDS dataset description model.
converter = make_converter()

converter.register_unstructure_hook(
    BIDSDatasetDescription,
    make_dict_unstructure_fn(
        BIDSDatasetDescription,
        converter,
        **{
            a.name: override(rename=_rename(a.name))
            for a in fields(BIDSDatasetDescription)
        },
    ),
)



from os import PathLike

from pydra import Submitter, Workflow
from pydra.engine.specs import Result


def list_out_fields(wf: Workflow) -> list:
    """Extract the output fields from a Workflow

    Parameters
    ----------
    wf : Workflow

    Returns
    -------
    list
        list of wf workflow's fields from output_spec
    """
    return [x[0] for x in wf.output_spec.fields if not x[0].startswith("_")]


def list_workflow_inputs(wf: Workflow) -> dict:
    """Extract the input default dictionary values from a Workflow.

    This is used as a way to pass data from the core Workflow
    to the input_workflow in order to properly query input folders.

    Parameters
    ----------
    wf : Workflow
        Workflow to list inputs from.

    Returns
    -------
    dict : Dictionary of wf default values from input_spec.
    """
    inputs = {}
    for input_field in wf.input_spec.fields:
        if not input_field[0].startswith("_"):
            if input_field[1] == dict:
                inputs[input_field[0]] = input_field[2]
            elif input_field[1] == str:
                inputs[input_field[0]] = {}
            else:
                raise ValueError(
                    f"Could not parse input field {input_field}."
                    "Please review your input specifications."
                )
    return inputs


def run(wf: Workflow) -> Result:
    """Execute a Pydra workflow.

    If the execution of the workflow fails, the
    errors are printed on stdout.

    Parameters
    ----------
    wf : Workflow
        The workflow to be executed.

    Returns
    -------
    Result :
        The result of running the Workflow

    """
    import re

    try:
        with Submitter(plugin="cf") as submitter:
            submitter(wf)
    except Exception as e:
        path = re.search(r"/.*\.pklz", str(e))
        if path:
            print(read_error(path.group(0)))
        print(str(e))

    return wf.result(return_inputs=False)


def read_error(path: PathLike) -> str:
    """Read pklz file with error message

    Parameters
    ----------
    path : PathLike
        The path of the file containing the error message

    Returns
    -------
    str
        The error message
    """

    import cloudpickle as cp

    with open(path, "rb") as fp:
        err = cp.load(fp)
    return err["error message"][:-1]



"""This module contains utilities to grab or download files for Clinica."""

import hashlib
import os
from collections import namedtuple
from pathlib import Path
from typing import Dict, List, Optional, Tuple

RemoteFileStructure = namedtuple("RemoteFileStructure", ["filename", "url", "checksum"])


def insensitive_glob(pattern_glob: str, recursive: Optional[bool] = False) -> List[str]:
    """This function is the glob.glob() function that is insensitive to the case.

    Parameters
    ----------
    pattern_glob : str
        Sensitive-to-the-case pattern.

    recursive : bool, optional
        Recursive parameter for `glob.glob()`.
        Default=False.

    Returns
    -------
    List[str] :
        Insensitive-to-the-case pattern.
    """
    from glob import glob

    def either(c: str) -> str:
        return "[%s%s]" % (c.lower(), c.upper()) if c.isalpha() else c

    return glob("".join(map(either, pattern_glob)), recursive=recursive)


def determine_caps_or_bids(input_dir: os.PathLike) -> bool:
    """Determine if the `input_dir` is a CAPS or a BIDS folder.

    Parameters
    ----------
    input_dir : os.PathLike
        The input folder.

    Returns
    -------
    bool :
        True if `input_dir` is a BIDS folder, False if `input_dir`
        is a CAPS folder or could not be determined.
    """
    input_dir = Path(input_dir)
    subjects_dir = input_dir / "subjects"
    groups_dir = input_dir / "groups"

    dir_to_look = subjects_dir if subjects_dir.is_dir() else input_dir
    subjects_sub_folders = _list_subjects_sub_folders(dir_to_look, groups_dir)
    if subjects_dir.is_dir():
        return False
    return len(subjects_sub_folders) > 0


def _list_subjects_sub_folders(
    root_dir: os.PathLike, groups_dir: os.PathLike
) -> List[os.PathLike]:
    from clinica.utils.stream import cprint

    root_dir = Path(root_dir)
    groups_dir = Path(groups_dir)
    warning_msg = (
        f"Could not determine if {groups_dir.parent} is a CAPS or BIDS directory. "
        "Clinica will assume this is a CAPS directory."
    )
    folder_content = [f for f in root_dir.iterdir()]
    subjects_sub_folders = [
        sub
        for sub in folder_content
        if (sub.name.startswith("sub-") and (root_dir / sub).is_dir())
    ]
    if len(subjects_sub_folders) == 0 and not groups_dir.is_dir():
        cprint(msg=warning_msg, lvl="warning")
    return subjects_sub_folders


def _common_checks(directory: os.PathLike, folder_type: str) -> None:
    """Utility function which performs checks common to BIDS and CAPS folder structures.

    Parameters
    ----------
    directory : PathLike
        Directory to check.

    folder_type : {"BIDS", "CAPS"}
        The type of directory.
    """
    from clinica.utils.exceptions import ClinicaBIDSError, ClinicaCAPSError

    if not isinstance(directory, (os.PathLike, str)):
        raise ValueError(
            f"Argument you provided to check_{folder_type.lower()}_folder() is not a string."
        )

    error = ClinicaBIDSError if folder_type == "BIDS" else ClinicaCAPSError

    if not os.path.isdir(directory):
        raise error(
            f"The {folder_type} directory you gave is not a folder.\n"
            "Error explanations:\n"
            f"\t- Clinica expected the following path to be a folder: {directory}\n"
            "\t- If you gave relative path, did you run Clinica on the good folder?"
        )


def check_bids_folder(bids_directory: os.PathLike) -> None:
    """Check if provided `bids_directory` is a BIDS folder.

    Parameters
    ----------
    bids_directory : PathLike
        The input folder to check.

    Raises
    ------
    ValueError :
        If `bids_directory` is not a string.

    ClinicaBIDSError :
        If the provided path does not exist, or is not a directory.
        If the provided path is a CAPS folder (BIDS and CAPS could
        be swapped by user). We simply check that there is not a folder
        called 'subjects' in the provided path (that exists in CAPS hierarchy).
        If the provided folder is empty.
        If the provided folder does not contain at least one directory whose
        name starts with 'sub-'.
    """
    from clinica.utils.exceptions import ClinicaBIDSError

    bids_directory = Path(bids_directory)
    _common_checks(bids_directory, "BIDS")

    if (bids_directory / "subjects").is_dir():
        raise ClinicaBIDSError(
            f"The BIDS directory ({bids_directory}) you provided seems to "
            "be a CAPS directory due to the presence of a 'subjects' folder."
        )

    if len([f for f in bids_directory.iterdir()]) == 0:
        raise ClinicaBIDSError(
            f"The BIDS directory you provided is empty. ({bids_directory})."
        )

    subj = [f for f in bids_directory.iterdir() if f.name.startswith("sub-")]
    if len(subj) == 0:
        raise ClinicaBIDSError(
            "Your BIDS directory does not contains a single folder whose name "
            "starts with 'sub-'. Check that your folder follow BIDS standard."
        )


def check_caps_folder(caps_directory: os.PathLike) -> None:
    """Check if provided `caps_directory`is a CAPS folder.

    Parameters
    ----------
    caps_directory : os.PathLike
        The input folder to check.

    Raises
    ------
    ValueError :
        If `caps_directory` is not a string.

    ClinicaCAPSError :
        If the provided path does not exist, or is not a directory.
        If the provided path is a BIDS folder (BIDS and CAPS could be
        swapped by user). We simply check that there is not a folder
        whose name starts with 'sub-' in the provided path (that exists
        in BIDS hierarchy).

    Notes
    -----
    Keep in mind that a CAPS folder can be empty.
    """
    from clinica.utils.exceptions import ClinicaCAPSError

    caps_directory = Path(caps_directory)
    _common_checks(caps_directory, "CAPS")

    sub_folders = [f for f in caps_directory.iterdir() if f.name.startswith("sub-")]
    if len(sub_folders) > 0:
        error_string = (
            "Your CAPS directory contains at least one folder whose name "
            "starts with 'sub-'. Check that you did not swap BIDS and CAPS folders.\n"
            "Folder(s) found that match(es) BIDS architecture:\n"
        )
        for directory in sub_folders:
            error_string += f"\t{directory}\n"
        error_string += (
            "A CAPS directory has a folder 'subjects' at its root, in which "
            "are stored the output of the pipeline for each subject."
        )
        raise ClinicaCAPSError(error_string)


def find_sub_ses_pattern_path(
    input_directory: os.PathLike,
    subject: str,
    session: str,
    error_encountered: list,
    results: list,
    is_bids: bool,
    pattern: str,
) -> None:
    """Appends the output path corresponding to subject, session and pattern in results.

    If an error is encountered, its corresponding message is added to the list `error_encountered`.

    Parameters
    ----------
    input_directory : str
        Path to the root of the input directory (BIDS or CAPS).

        .. warning::
            This function does not perform any check on `input_directory`.
            It is assumed that it has been previously checked by either
            `check_bids_directory` or `check_caps_directory`, and that
            the flag `is_bids` has been set accordingly.

    subject : str
        Name given to the folder of a participant (ex: sub-ADNI002S0295).

    session : str
        Name given to the folder of a session (ex: ses-M00).

    error_encountered : List
        List to which errors encountered in this function are added.

    results : List
        List to which the output path corresponding to subject, session
        and pattern is added.

    is_bids : bool
        True if `input_dir` is a BIDS folder, False if `input_dir` is a
        CAPS folder.

    pattern : str
        Define the pattern of the final file.
    """
    input_directory = Path(input_directory)
    if is_bids:
        origin_pattern = input_directory / subject / session
    else:
        origin_pattern = input_directory / "subjects" / subject / session

    current_pattern = origin_pattern / "**" / pattern
    current_glob_found = insensitive_glob(str(current_pattern), recursive=True)
    # Error handling if more than 1 file are found, or when no file is found
    if len(current_glob_found) > 1:
        error_str = f"\t*  ({subject} | {session}): More than 1 file found:\n"
        for found_file in current_glob_found:
            error_str += f"\t\t{found_file}\n"
        error_encountered.append(error_str)
    elif len(current_glob_found) == 0:
        error_encountered.append(f"\t* ({subject} | {session}): No file found\n")
    # Otherwise the file found is added to the result
    else:
        results.append(current_glob_found[0])


def _check_information(information: Dict) -> None:
    if not isinstance(information, (dict, list)):
        raise TypeError(
            "A dict or list of dicts must be provided for the argument 'information'"
        )

    if isinstance(information, list):
        for item in information:
            if not all(elem in item for elem in ["pattern", "description"]):
                raise ValueError(
                    "'information' must contain the keys 'pattern' and 'description'"
                )

            if not all(
                elem in ["pattern", "description", "needed_pipeline"]
                for elem in item.keys()
            ):
                raise ValueError(
                    "'information' can only contain the keys 'pattern', 'description' and 'needed_pipeline'"
                )

            if item["pattern"][0] == "/":
                raise ValueError(
                    "pattern argument cannot start with char: / (does not work in os.path.join function). "
                    "If you want to indicate the exact name of the file, use the format "
                    "directory_name/filename.extension or filename.extension in the pattern argument."
                )
    else:
        if not all(elem in information for elem in ["pattern", "description"]):
            raise ValueError(
                "'information' must contain the keys 'pattern' and 'description'"
            )

        if not all(
            elem in ["pattern", "description", "needed_pipeline"]
            for elem in information.keys()
        ):
            raise ValueError(
                "'information' can only contain the keys 'pattern', 'description' and 'needed_pipeline'"
            )

        if information["pattern"][0] == "/":
            raise ValueError(
                "pattern argument cannot start with char: / (does not work in os.path.join function). "
                "If you want to indicate the exact name of the file, use the format "
                "directory_name/filename.extension or filename.extension in the pattern argument."
            )


def _format_errors(errors: List, information: Dict) -> str:
    error_message = (
        f"Clinica encountered {len(errors)} "
        f"problem(s) while getting {information['description']}:\n"
    )
    if "needed_pipeline" in information and information["needed_pipeline"]:
        error_message += (
            "Please note that the following clinica pipeline(s) must "
            f"have run to obtain these files: {information['needed_pipeline']}\n"
        )
    error_message += "\n".join(errors)

    return error_message


def clinica_file_reader(
    subjects: List[str],
    sessions: List[str],
    input_directory: os.PathLike,
    information: Dict,
    raise_exception: Optional[bool] = True,
    n_procs: Optional[int] = 1,
):
    """Read files in BIDS or CAPS directory based on participant ID(s).

    This function grabs files relative to a subject and session list according to a glob pattern (using *)

    Parameters
    ----------
    subjects : List[str]
        List of subjects.

    sessions : List[str]
        List of sessions. Must be same size as `subjects` and must correspond.

    input_directory : PathLike
        Path to the BIDS or CAPS directory to read from.

    information : Dict
        Dictionary containing all the relevant information to look for the files.
        The possible keys are:

            - `pattern`: Required. Define the pattern of the final file.
            - `description`: Required. String to describe what the file is.
            - `needed_pipeline` : Optional. String describing the pipeline(s)
              needed to obtain the related file.

    raise_exception : bool, optional
        If True, an exception is raised if errors happen. If not, we return the file
        list as it is. Default=True.

    n_procs : int, optional
        Number of cores used to fetch files in parallel.
        If set to 1, subjects and sessions will be processed sequentially.
        Default=1.

    Returns
    -------
    results : List[str]
        List of files respecting the subject/session order provided in input.

    error_message : str
        Error message which contains all errors encountered while reading the files.

    Raises
    ------
    TypeError
        If `information` is not a dictionary.

    ValueError
        If `information` is not formatted correctly. See function `_check_information`
        for more details.
        If the length of `subjects` is different from the length of `sessions`.

    ClinicaCAPSError or ClinicaBIDSError
        If multiples files are found for 1 subject/session, or if no file is found.

        .. note::
            If `raise_exception` is False, no exception is raised.

    Notes
    -----
    This function is case-insensitive, meaning that the pattern argument can, for example,
    contain upper case letters that do not exist in the existing file path.

    You should always use `clinica_file_reader` in the following manner:

    .. code-block:: python

         try:
            file_list = clinica_file_reader(...)
         except ClinicaException as e:
            # Deal with the error

    Examples
    --------
    The paths are shortened for readability.

    You have the full name of a file.

    File `orig_nu.mgz` from FreeSurfer of subject `sub-ADNI011S4105`, session `ses-M00`
    located in mri folder of FreeSurfer output :

    >>> clinica_file_reader(
            ['sub-ADNI011S4105'],
            ['ses-M00'],
            caps_directory,
            {
                'pattern': 'freesurfer_cross_sectional/sub-*_ses-*/mri/orig_nu.mgz',
                'description': 'freesurfer file orig_nu.mgz',
                'needed_pipeline': 't1-freesurfer'
            }
        )
    ['/caps/subjects/sub-ADNI011S4105/ses-M00/t1/freesurfer_cross_sectional/sub-ADNI011S4105_ses-M00/mri/orig_nu.mgz']

    You have a partial name of the file.

    File `sub-ADNI011S4105_ses-M00_trc-18FFDG_pet.nii.gz` in BIDS directory.
    Here, filename depends on subject and session name :

    >>> clinica_file_reader(
            ['sub-ADNI011S4105'],
            ['ses-M00'],
            bids_directory,
            {
                'pattern': '*18FFDG_pet.nii*',
                'description': 'FDG PET data'
            }
        )
    ['/bids/sub-ADNI011S4105/ses-M00/pet/sub-ADNI011S4105_ses-M00_trc-18FFDG_pet.nii.gz']

    Tricky example.

    Get the file `rh.white` from FreeSurfer :

    This will fail :

    >>> clinica_file_reader(
            ['sub-ADNI011S4105'],
            ['ses-M00'],
            caps,
            {
                'pattern': 'rh.white',
                'description': 'right hemisphere of outter cortical surface.',
                'needed_pipeline': 't1-freesurfer'
            }
        )
    * More than 1 file found::
            /caps/subjects/sub-ADNI011S4105/ses-M00/t1/freesurfer_cross_sectional/fsaverage/surf/rh.white
            /caps/subjects/sub-ADNI011S4105/ses-M00/t1/freesurfer_cross_sectional/rh.EC_average/surf/rh.white
            /caps/subjects/sub-ADNI011S4105/ses-M00/t1/freesurfer_cross_sectional/sub-ADNI011S4105_ses-M00/surf/rh.white

    Correct usage (e.g. in pet-surface): pattern string must be 'sub-*_ses-*/surf/rh.white',
    or even more precise: 't1/freesurfer_cross_sectional/sub-*_ses-*/surf/rh.white'
    It then gives: ['/caps/subjects/sub-ADNI011S4105/ses-M00/t1/freesurfer_cross_sectional/sub-ADNI011S4105_ses-M00/surf/rh.white']
    """
    from clinica.utils.exceptions import ClinicaBIDSError, ClinicaCAPSError

    _check_information(information)
    pattern = information["pattern"]

    input_directory = Path(input_directory)
    is_bids = determine_caps_or_bids(input_directory)
    if is_bids:
        check_bids_folder(input_directory)
    else:
        check_caps_folder(input_directory)

    if len(subjects) != len(sessions):
        raise ValueError("Subjects and sessions must have the same length.")

    if len(subjects) == 0:
        return [], ""

    file_reader = _read_files_parallel if n_procs > 1 else _read_files_sequential
    results, errors_encountered = file_reader(
        input_directory,
        subjects,
        sessions,
        is_bids,
        pattern,
        n_procs=n_procs,
    )
    error_message = _format_errors(errors_encountered, information)

    if len(errors_encountered) > 0 and raise_exception:
        if is_bids:
            raise ClinicaBIDSError(error_message)
        else:
            raise ClinicaCAPSError(error_message)

    return results, error_message


def _read_files_parallel(
    input_directory: os.PathLike,
    subjects: List[str],
    sessions: List[str],
    is_bids: bool,
    pattern: str,
    n_procs: int,
) -> Tuple[List[str], List[str]]:
    from multiprocessing import Manager

    from joblib import Parallel, delayed

    manager = Manager()
    shared_results = manager.list()
    shared_errors_encountered = manager.list()
    Parallel(n_jobs=n_procs)(
        delayed(find_sub_ses_pattern_path)(
            input_directory,
            sub,
            ses,
            shared_errors_encountered,
            shared_results,
            is_bids,
            pattern,
        )
        for sub, ses in zip(subjects, sessions)
    )
    results = list(shared_results)
    errors_encountered = list(shared_errors_encountered)
    return results, errors_encountered


def _read_files_sequential(
    input_directory: os.PathLike,
    subjects: List[str],
    sessions: List[str],
    is_bids: bool,
    pattern: str,
    **kwargs,
) -> Tuple[List[str], List[str]]:
    errors_encountered, results = [], []
    for sub, ses in zip(subjects, sessions):
        find_sub_ses_pattern_path(
            input_directory, sub, ses, errors_encountered, results, is_bids, pattern
        )
    return results, errors_encountered


def clinica_list_of_files_reader(
    participant_ids: List[str],
    session_ids: List[str],
    bids_or_caps_directory: os.PathLike,
    list_information: List[Dict],
    raise_exception: Optional[bool] = True,
) -> List[List[str]]:
    """Read list of BIDS or CAPS files.

    This function iterates calls of `clinica_file_reader` to extract input files based
    on information given by `list_information`.

    Parameters
    ----------
    participant_ids : List[str]
        List of participant IDs.
        Example: ['sub-CLNC01', 'sub-CLNC01', 'sub-CLNC02']

    session_ids : List[str]
        List of sessions ID associated to `participant_ids`
        Example: ['ses-M00', 'ses-M18', 'ses-M00']

    bids_or_caps_directory : PathLike
        Path to the BIDS of CAPS directory to read from.

    list_information : List[Dict]
        List of information dictionaries described in `clinica_file_reader`.

    raise_exception : bool, optional
        Raise Exception or not. Defaults to True.

    Returns
    -------
    list_found_files : List[List[str]]
        List of lists of found files following order of `list_information`
    """
    from .exceptions import ClinicaBIDSError, ClinicaException

    all_errors = []
    list_found_files = []
    for info_file in list_information:
        try:
            list_found_files.append(
                clinica_file_reader(
                    participant_ids,
                    session_ids,
                    bids_or_caps_directory,
                    info_file,
                    True,
                )[0]
            )
        except ClinicaException as e:
            list_found_files.append([])
            all_errors.append(e)

    if len(all_errors) > 0 and raise_exception:
        error_message = "Clinica faced error(s) while trying to read files in your BIDS or CAPS directory.\n"
        for msg in all_errors:
            error_message += str(msg)
        raise ClinicaBIDSError(error_message)

    return list_found_files


def clinica_group_reader(
    caps_directory: os.PathLike,
    information: Dict,
    raise_exception: Optional[bool] = True,
) -> str:
    """Read files from CAPS directory based on group ID(s).

    This function grabs files relative to a group, according to a glob pattern (using *).
    Only one file can be returned, as order is arbitrary in `glob.glob()`.

    Parameters
    ----------
    caps_directory : PathLike
        Path to the input CAPS directory.

    information : Dict
        Dictionary containing all the relevant information to look for the files.
        The possible keys are:

            - `pattern`: Required. Define the pattern of the final file.
            - `description`: Required. String to describe what the file is.
            - `needed_pipeline` : Optional. String describing the pipeline(s)
              needed to obtain the related file.

    raise_exception : bool, optional
        If True, an exception is raised if errors happen.
        If not, we return the file list as it is.
        Default=True.

    Returns
    -------
    str :
        Path to the found file.

    Raises
    ------
    ClinicaCAPSError :
        If no file is found, or more than 1 files are found.
    """
    _check_information(information)
    pattern = information["pattern"]
    caps_directory = Path(caps_directory)
    check_caps_folder(caps_directory)

    current_pattern = caps_directory / "**" / pattern
    found_files = insensitive_glob(str(current_pattern), recursive=True)

    # Since we are returning found_files[0], force raising even if raise_exception is False
    # Otherwise we'll get an uninformative IndexError...
    if (len(found_files) == 0) or (len(found_files) > 1 and raise_exception is True):
        _format_and_raise_group_reader_errors(caps_directory, found_files, information)

    return found_files[0]


def _format_and_raise_group_reader_errors(
    caps_directory: os.PathLike,
    found_files: List,
    information: Dict,
) -> None:
    from clinica.utils.exceptions import ClinicaCAPSError

    error_string = (
        f"Clinica encountered a problem while getting {information['description']}. "
    )
    if len(found_files) == 0:
        error_string += "No file was found"
    else:
        error_string += f"{len(found_files)} files were found:"
        for found_file in found_files:
            error_string += f"\n\t{found_file}"
        error_string += (
            f"\n\tCAPS directory: {caps_directory}\n"
            "Please note that the following clinica pipeline(s) must have run to obtain these files: "
            f"{information['needed_pipeline']}\n"
        )
    raise ClinicaCAPSError(error_string)


def _sha256(path):
    """Calculate the sha256 hash of the file at path."""
    sha256hash = hashlib.sha256()
    chunk_size = 8192
    with open(path, "rb") as f:
        while True:
            buffer = f.read(chunk_size)
            if not buffer:
                break
            sha256hash.update(buffer)
    return sha256hash.hexdigest()


def fetch_file(remote: RemoteFileStructure, dirname: Optional[str]) -> str:
    """Download a specific file and save it into the resources folder of the package.

    Parameters
    ----------
    remote : RemoteFileStructure
        Structure containing url, filename and checksum.

    dirname : str
        Absolute path where the file will be downloaded.

    Returns
    -------
    file_path : str
        Absolute file path.
    """
    import os.path
    import shutil
    import ssl
    from urllib.error import URLError
    from urllib.request import Request, urlopen

    from clinica.utils.stream import cprint

    if not os.path.exists(dirname):
        cprint(msg="Path to the file does not exist", lvl="warning")
        cprint(msg="Stop Clinica and handle this error", lvl="warning")

    file_path = os.path.join(dirname, remote.filename)
    # Download the file from `url` and save it locally under `file_name`:
    gcontext = ssl.SSLContext()
    req = Request(remote.url + remote.filename)
    try:
        response = urlopen(req, context=gcontext)
    except URLError as e:
        if hasattr(e, "reason"):
            cprint(msg=f"We failed to reach a server. Reason: {e.reason}", lvl="error")
        elif hasattr(e, "code"):
            cprint(
                msg=f"The server could not fulfill the request. Error code: {e.code}",
                lvl="error",
            )
    else:
        try:
            with open(file_path, "wb") as out_file:
                shutil.copyfileobj(response, out_file)
        except OSError as err:
            cprint(msg="OS error: {0}".format(err), lvl="error")

    checksum = _sha256(file_path)
    if remote.checksum != checksum:
        raise IOError(
            f"{file_path} has an SHA256 checksum ({checksum}) from expected "
            f"({remote.checksum}), file may be corrupted."
        )
    return file_path


def get_file_from_server(
    remote_file: RemoteFileStructure,
    cache_path: Optional[str] = None,
) -> str:
    """Download file from server.

    Parameters
    ----------
    remote_file : RemoteFileStructure
        Remote file structure defined in `clinica.utils.inputs` specifying
        the file to retrieve.

    cache_path : str, optional
        Path to cache. Default="~/.cache/clinica/data".

    Returns
    -------
    local_file : str
        Path to the downloaded file.
    """
    import os
    from pathlib import Path

    from clinica.utils.stream import cprint

    home = str(Path.home())
    if cache_path:
        cache_clinica = os.path.join(home, ".cache", cache_path)
    else:
        cache_clinica = os.path.join(home, ".cache", "clinica", "data")

    os.makedirs(cache_clinica, exist_ok=True)

    local_file = os.path.join(cache_clinica, remote_file.filename)

    if not (os.path.exists(local_file)):
        try:
            local_file = fetch_file(remote_file, cache_clinica)
        except IOError as err:
            cprint(
                msg=f"Unable to download {remote_file.filename} from {remote_file.url}: {err}",
                lvl="error",
            )

    return local_file
